part of datagrid;

class GridCell extends StatefulWidget {
  const GridCell({
    @required this.key,
    this.dataCell,
    this.padding,
    this.backgroundColor,
    this.isDirty,
    this.child,
    this.alignment,
  }) : super(key: key);

  final Key key;
  final DataCellBase dataCell;
  final Widget child;
  final EdgeInsets padding;
  final Color backgroundColor;
  final bool isDirty;
  final Alignment alignment;

  @override
  State<StatefulWidget> createState() => _GridCellState();
}

class _GridCellState extends State<GridCell> {
  @override
  Widget build(BuildContext context) {
    final Widget child = Semantics(
      label: widget.dataCell.cellValue.toString(),
      child: _wrapInsideGestureDetector(),
    );

    return _GridCellRenderObjectWidget(
      key: widget.key,
      child: child,
      dataCell: widget.dataCell,
      isDirty: widget.isDirty,
    );
  }

  Widget _wrapInsideGestureDetector() {
    return GestureDetector(
      onTapUp: _handleOnTapUp,
      child: this._wrapInsideContainer(),
    );
  }

  Widget _wrapInsideContainer() {
    return Container(
        key: widget.key,
        clipBehavior: Clip.antiAlias,
        decoration: BoxDecoration(border: _getCellBorder(widget.dataCell)),
        alignment: Alignment.center,
        child: _wrapInsideCellContainer(widget, widget.child));
  }

  void _handleOnTapUp(TapUpDetails e) {
    final _DataGridSettings dataGridSettings =
        widget.dataCell?._dataRow?._dataGridStateDetails();
    if (dataGridSettings == null || widget.dataCell == null) return;

    dataGridSettings.dataGridFocusNode.requestFocus();
    widget.dataCell._onTouchUp();
  }
}

class GridWidgetCell extends GridCell {
  const GridWidgetCell({
    @required this.key,
    this.dataCell,
    this.padding,
    this.backgroundColor,
    this.isDirty,
    this.child,
  }) : super(key: key);

  final Key key;
  final DataCellBase dataCell;
  final Widget child;
  final EdgeInsets padding;
  final Color backgroundColor;
  final bool isDirty;

  @override
  State<StatefulWidget> createState() => _GridWidgetCellState();
}

class _GridWidgetCellState extends State<GridWidgetCell> {
  @override
  Widget build(BuildContext context) {
    final Widget child = this._wrapInsideGestureDetector(context);

    return _GridCellRenderObjectWidget(
      key: widget.key,
      dataCell: widget.dataCell,
      child: child,
      isDirty: widget.isDirty,
    );
  }

  Widget _wrapInsideGestureDetector(BuildContext context) {
    return GestureDetector(
      onTapUp: _handleOnTapUp,
      child: this._wrapInsideContainer(context),
    );
  }

  Widget _wrapInsideContainer(BuildContext context) {
    final _DataGridSettings dataGridSettings =
        widget.dataCell?._dataRow?._dataGridStateDetails();
    Widget child;
    if (dataGridSettings != null && dataGridSettings.cellBuilder != null) {
      var recordIndex = _GridIndexResolver.resolveToRecordIndex(
          dataGridSettings, widget.dataCell.rowIndex);
      child = dataGridSettings.cellBuilder(
          context, widget.dataCell.gridColumn, recordIndex);
    }

    return Container(
        key: widget.key,
        clipBehavior: Clip.antiAlias,
        decoration: BoxDecoration(border: _getCellBorder(widget.dataCell)),
        child: _wrapInsideCellContainer(widget, child));
  }

  void _handleOnTapUp(TapUpDetails e) {
    final _DataGridSettings dataGridSettings =
        widget.dataCell?._dataRow?._dataGridStateDetails();
    if (dataGridSettings == null || widget.dataCell == null) return;

    dataGridSettings.dataGridFocusNode.requestFocus();
    widget.dataCell._onTouchUp();
  }
}

class _GridCellRenderObjectWidget extends SingleChildRenderObjectWidget {
  _GridCellRenderObjectWidget(
      {@required this.key, this.dataCell, this.isDirty, this.child})
      : super(key: key, child: RepaintBoundary.wrap(child, 0));

  final Widget child;
  final Key key;
  final DataCellBase dataCell;
  final bool isDirty;

  @override
  _RenderGridCell createRenderObject(BuildContext context) {
    return _RenderGridCell(
      dataCell: this.dataCell,
      isDirty: this.isDirty,
    );
  }

  @override
  void updateRenderObject(BuildContext context, _RenderGridCell renderObject) {
    super.updateRenderObject(context, renderObject);
    renderObject
      ..isDirty = this.isDirty
      ..dataCell = this.dataCell;
  }
}

class _RenderGridCell extends RenderBox
    with RenderObjectWithChildMixin<RenderBox> {
  DataCellBase get dataCell => this._dataCell;
  DataCellBase _dataCell;

  set dataCell(DataCellBase newDataColumn) {
    if (this._dataCell == newDataColumn) return;

    _dataCell = newDataColumn;
    this.markNeedsLayout();
    this.markNeedsPaint();
  }

  bool get isDirty => this._isDirty;
  bool _isDirty = false;

  set isDirty(bool newValue) {
    this._isDirty = newValue;
    if (this._isDirty) {
      this.markNeedsLayout();
      this.markNeedsPaint();
    }

    this.dataCell?._isDirty = false;
  }

  Rect get columnRect => this._columnRect;
  Rect _columnRect = Rect.zero;

  set columnRect(Rect columnRect) {
    _columnRect = columnRect;
  }

  _RenderGridCell({RenderBox child, DataCellBase dataCell, bool isDirty})
      : _dataCell = dataCell,
        _isDirty = isDirty;

  Rect _measureColumnRect(double rowHeight) {
    if (this.dataCell._dataRow.isVisible && this.dataCell.isVisible) {
      final DataRow dataRow = this.dataCell._dataRow;
      final _DataGridSettings _dataGrid = dataRow._dataGridStateDetails();
      final _VisualContainerHelper container = _dataGrid.container;
      final _VisibleLineInfo lineInfo =
          dataRow._getColumnVisibleLineInfo(dataCell.columnIndex);
      final double lineSize =
          this.dataCell._dataRow._getColumnSize(dataCell.columnIndex, false);

      var origin = lineInfo != null ? lineInfo.origin : 0.0;
      origin += container.horizontalOffset;

      //To overcome grid common RightToLeft clipping line creation problem instead of handling in grid common source.
      if (_dataGrid.textDirection == TextDirection.rtl &&
          lineInfo != null &&
          lineInfo.visibleIndex ==
              _SfDataGridHelper.getVisibleLines(dataRow._dataGridStateDetails())
                  .firstBodyVisibleIndex) {
        origin += lineInfo.scrollOffset;
      }

      this.columnRect = Rect.fromLTWH(origin, 0, lineSize, rowHeight);
      origin = null;
    } else {
      this.columnRect = Rect.zero;
    }

    return this.columnRect;
  }

  @override
  bool hitTestChildren(BoxHitTestResult result, {Offset position}) {
    if (this.child == null) return false;

    final BoxParentData childParentData =
        this.child.parentData as BoxParentData;
    final bool isHit = result.addWithPaintOffset(
      offset: childParentData.offset,
      position: position,
      hitTest: (BoxHitTestResult result, Offset transformed) {
        return child.hitTest(result, position: transformed);
      },
    );
    if (isHit) {
      return true;
    } else {
      return false;
    }
  }

  @override
  bool hitTestSelf(Offset position) => true;

  @override
  bool get isRepaintBoundary => true;

  @override
  void performLayout() {
    size = constraints
        .constrain(Size(this.constraints.maxWidth, this.constraints.maxHeight));

    if (this.child != null) {
      this.child.layout(
          BoxConstraints.tightFor(
              width: this.constraints.maxWidth,
              height: this.constraints.maxHeight),
          parentUsesSize: true);
    }
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    if (this.child != null) {
      context.paintChild(this.child, offset);
    }

    super.paint(context, offset);
  }
}

BorderDirectional _getCellBorder(DataCellBase dataCell) {
  final _DataGridSettings dataGridSettings =
      dataCell._dataRow._dataGridStateDetails();
  final Color borderColor = dataGridSettings.dataGridThemeData?.gridLineColor;
  final borderWidth = dataGridSettings.dataGridThemeData?.gridLineStrokeWidth;

  final bool canDrawHorizontalBorder =
      dataGridSettings.gridLinesVisibility == GridLinesVisibility.horizontal ||
          dataGridSettings.gridLinesVisibility == GridLinesVisibility.both;

  final bool canDrawVerticalBorder =
      dataGridSettings.gridLinesVisibility == GridLinesVisibility.vertical ||
          dataGridSettings.gridLinesVisibility == GridLinesVisibility.both;

  _getLeftBorder() {
    if ((dataCell.columnIndex == 0 && canDrawVerticalBorder))
      return BorderSide(width: borderWidth, color: borderColor);
    else
      return BorderSide.none;
  }

  _getTopBorder() {
    if ((dataCell.rowIndex == 0 && canDrawHorizontalBorder))
      return BorderSide(width: borderWidth, color: borderColor);
    else
      return BorderSide.none;
  }

  _getRightBorder() {
    if (canDrawVerticalBorder)
      return BorderSide(width: borderWidth, color: borderColor);
    else
      return BorderSide.none;
  }

  _getBottomBorder() {
    if (canDrawHorizontalBorder)
      return BorderSide(width: borderWidth, color: borderColor);
    else
      return BorderSide.none;
  }

  return BorderDirectional(
    start: _getLeftBorder(),
    top: _getTopBorder(),
    end: _getRightBorder(),
    bottom: _getBottomBorder(),
  );
}

Widget _wrapInsideCellContainer(GridCell widget, Widget child) {
  var dataGridSettings = widget.dataCell._dataRow._dataGridStateDetails();
  if (dataGridSettings == null) return child;

  final color = dataGridSettings.dataGridThemeData.currentCellStyle.borderColor;
  final borderWidth =
      dataGridSettings.dataGridThemeData.currentCellStyle.borderWidth ?? 1.0;

  getBorder() {
    var isCurrentCell = widget.dataCell.isCurrentCell;
    return Border(
      bottom: isCurrentCell
          ? BorderSide(color: color, width: borderWidth)
          : BorderSide.none,
      left: isCurrentCell
          ? BorderSide(color: color, width: borderWidth)
          : BorderSide.none,
      top: isCurrentCell
          ? BorderSide(color: color, width: borderWidth)
          : BorderSide.none,
      right: isCurrentCell
          ? BorderSide(color: color, width: borderWidth)
          : BorderSide.none,
    );
  }

  return LayoutBuilder(
    builder: (context, constraint) {
      return Container(
          key: widget.key,
          width: constraint.maxWidth,
          height: constraint.maxHeight,
          padding: _getPadding(widget.dataCell, widget.padding),
          alignment: widget.alignment,
          clipBehavior: Clip.antiAlias,
          decoration:
              BoxDecoration(color: widget.backgroundColor, border: getBorder()),
          child: child);
    },
  );
}

EdgeInsets _getPadding(DataCellBase dataCell, EdgeInsets padding) {
  var dataGridSettings = dataCell._dataRow?._dataGridStateDetails();

  if (dataGridSettings == null) return EdgeInsets.all(0.0);
  padding = padding != null
      ? padding -
          _getMeasuredStrokeWidthPadding(
              dataGridSettings, dataCell, dataCell.isCurrentCell)
      : padding;
  return padding != null && padding.isNonNegative ? padding : EdgeInsets.zero;
}

EdgeInsets _getMeasuredStrokeWidthPadding(
    _DataGridSettings dataGridSettings, DataCell dataCell, bool isCurrentCell) {
  var rowIndex = dataCell.rowIndex;
  var themeData = dataGridSettings.dataGridThemeData;
  var currentCellStrokeWidth = themeData.currentCellStyle.borderWidth ?? 1.0;

  if (rowIndex == 0) {
    return _getGridLineStrokeWidthPadding(
        dataGridSettings, dataCell, themeData?.gridLineStrokeWidth);
  } else {
    return _getGridLineStrokeWidthPadding(
            dataGridSettings, dataCell, themeData?.gridLineStrokeWidth) +
        EdgeInsets.all((isCurrentCell ? currentCellStrokeWidth : 0.0));
  }
}

EdgeInsets _getGridLineStrokeWidthPadding(_DataGridSettings dataGridSettings,
    DataCell dataCell, double gridLineStokeWidth) {
  final bool canConsiderHorizontalPadding =
      dataGridSettings.gridLinesVisibility == GridLinesVisibility.horizontal ||
          dataGridSettings.gridLinesVisibility == GridLinesVisibility.both;

  final bool canConsiderVerticalPadding =
      dataGridSettings.gridLinesVisibility == GridLinesVisibility.vertical ||
          dataGridSettings.gridLinesVisibility == GridLinesVisibility.both;

  _getLeftPadding() {
    if ((dataCell.columnIndex == 0 && canConsiderVerticalPadding))
      return gridLineStokeWidth;
    else
      return 0.0;
  }

  _getTopPadding() {
    if ((dataCell.rowIndex == 0 && canConsiderHorizontalPadding))
      return gridLineStokeWidth;
    else
      return 0.0;
  }

  _getRightPadding() {
    if (canConsiderVerticalPadding)
      return gridLineStokeWidth;
    else
      return 0.0;
  }

  _getBottomPadding() {
    if (canConsiderHorizontalPadding)
      return gridLineStokeWidth;
    else
      return 0.0;
  }

  return EdgeInsets.fromLTRB(_getLeftPadding(), _getTopPadding(),
      _getRightPadding(), _getBottomPadding());
}
