part of datagrid;

typedef _DataGridSourceListener = void Function(
    {RowColumnIndex rowColumnIndex, String propertyName});

/// A datasource for obtaining the row data for the [SfDataGrid].
///
/// The following APIs are mandatory to process the data,
/// * [dataSource] - The number of rows in a datagrid and row selection depends on the [dataSource]. So, set the collection required for datagrid in [dataSource].
/// * [getCellValue] - The data needed for the cells is obtained from [getCellValue].
///
/// Call the [notifyDataSourceListeners] when performing CRUD in the underlying datasource. When ///     updating data in a cell, pass the corresponding [RowColumnIndex] to the /// [notifyDataSourceListeners].
/// [DataGridSource] objects are expected to be long-lived, not recreated with each build.
/// ``` dart
/// final List<Employee> _employees = <Employee>[];
///
/// class EmployeeDataSource extends DataGridSource
/// {
///   @override
///   List<Object> get dataSource => _employees;
///
///   @override
///   getCellValue(int rowIndex, String columnName){
///     switch (columnName) {
///       case 'id':
///         return employees[rowIndex].id;
///         break;
///       case 'name':
///         return employees[rowIndex].name;
///         break;
///       case 'salary':
///         return employees[rowIndex].salary;
///         break;
///       case 'designation':
///         return employees[rowIndex].designation;
///         break;
///       default:
///         return ' ';
///         break;
///     }
///   }
/// ```
abstract class DataGridSource extends _DataGridSourceChangeNotifier {
  DataGridSource({List<Object> dataSource})
      : this._dataSource = dataSource ?? [];

  List<Object> get dataSource => this._dataSource;
  List<Object> _dataSource = [];

  getCellValue(int rowIndex, String columnName);

  @override
  bool operator ==(other) =>
      identical(this, other) ||
      other is DataGridSource && runtimeType == other.runtimeType;

  @override
  int get hashCode => super.hashCode;
}

class DataGridController extends _DataGridSourceChangeNotifier {
  _DataGridStateDetails _dataGridStateDetails;

  DataGridController(
      {int selectedIndex, Object selectedRow, List<Object> selectedRows})
      : this._selectedRow = selectedRow,
        this._selectedIndex = selectedIndex,
        this._selectedRows = selectedRows ?? [];

  List<Object> get selectedRows => this._selectedRows;
  List<Object> _selectedRows = [];

  set selectedRows(List<Object> newSelectedRows) {
    if (this._selectedRows == newSelectedRows) return;

    this._selectedRows = newSelectedRows;
    this.notifyDataSourceListeners(propertyName: 'selectedRows');
  }

  int get selectedIndex => this._selectedIndex;
  int _selectedIndex;

  set selectedIndex(int newSelectedIndex) {
    if (this._selectedIndex == newSelectedIndex) return;

    this._selectedIndex = newSelectedIndex;
    this.notifyDataSourceListeners(propertyName: 'selectedIndex');
  }

  Object get selectedRow => this._selectedRow;
  Object _selectedRow;

  set selectedRow(Object newSelectedRow) {
    if (this._selectedRow == newSelectedRow) return;

    this._selectedRow = newSelectedRow;
    this.notifyDataSourceListeners(propertyName: 'selectedRow');
  }

  RowColumnIndex get currentCell => this._currentCell;
  RowColumnIndex _currentCell = RowColumnIndex(-1, -1);

  void moveCurrentCellTo(RowColumnIndex rowColumnIndex) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    if (dataGridSettings != null &&
        rowColumnIndex != null &&
        rowColumnIndex != RowColumnIndex(-1, -1) &&
        dataGridSettings.selectionMode != SelectionMode.none &&
        dataGridSettings.navigationMode != GridNavigationMode.row) {
      var rowIndex = _GridIndexResolver.resolveToRowIndex(
          dataGridSettings, rowColumnIndex.rowIndex);
      var columnIndex = _GridIndexResolver.resolveToGridVisibleColumnIndex(
          dataGridSettings, rowColumnIndex.columnIndex);

      var isRowSelectionController =
          dataGridSettings.rowSelectionManager is RowSelectionManager;
      if (isRowSelectionController) {
        (dataGridSettings.rowSelectionManager as RowSelectionManager)
            ._processSelectionAndCurrentCell(
                dataGridSettings, RowColumnIndex(rowIndex, columnIndex));
      }
    }
  }

  @override
  bool operator ==(other) =>
      identical(this, other) ||
      other is DataGridController && runtimeType == other.runtimeType;

  @override
  int get hashCode => super.hashCode;
}

class _DataGridSourceChangeNotifier {
  ObserverList<_DataGridSourceListener> _listeners =
      ObserverList<_DataGridSourceListener>();

  void addListener(listener) {
    this._listeners.add(listener);
  }

  // ignore: unused_element
  bool get _hasListeners => this._listeners.isNotEmpty;

  void notifyDataSourceListeners(
      {RowColumnIndex rowColumnIndex, String propertyName}) {
    for (var listener in this._listeners) {
      listener(rowColumnIndex: rowColumnIndex, propertyName: propertyName);
    }
  }

  void removeListener(listener) {
    this._listeners.remove(listener);
  }

  @mustCallSuper
  void dispose() {
    this._listeners = null;
  }
}
