part of datagrid;

class GridColumn {
  double _autoWidth = double.nan;

  ColumnWidthMode get columnWidthMode => this._columnWidthMode;
  ColumnWidthMode _columnWidthMode = ColumnWidthMode.none;

  set columnWidthMode(ColumnWidthMode newValue) =>
      this._columnWidthMode = newValue;

  String get mappingName => this._mappingName;
  String _mappingName;

  set mappingName(String newValue) {
    if (this._mappingName == newValue) return;

    this._mappingName = newValue;
  }

  String get cellType => this._cellType;
  String _cellType;

  Alignment get textAlignment => this._textAlignment;
  Alignment _textAlignment = Alignment.centerLeft;

  set textAlignment(Alignment value) => this._textAlignment = value;

  Alignment get headerTextAlignment => this._headerTextAlignment;
  Alignment _headerTextAlignment = Alignment.center;

  set headerTextAlignment(Alignment value) => this._headerTextAlignment = value;

  int get maxLines => this._maxLines;
  int _maxLines;

  set maxLines(int newValue) => this._maxLines = newValue;

  TextOverflow get overflow => this._overflow;
  TextOverflow _overflow;

  set overflow(TextOverflow newValue) => this._overflow = newValue;

  bool get softWrap => this._softWrap;
  bool _softWrap = false;

  set softWrap(bool value) => _softWrap = value;

  double get actualWidth => this._actualWidth;
  double _actualWidth = double.nan;

  String get headerText => this._headerText;
  String _headerText;

  set headerText(String newValue) => this._headerText = newValue;

  double get minimumWidth => this._minimumWidth;
  double _minimumWidth = double.nan;

  set minimumWidth(double newValue) => this._minimumWidth = newValue;

  double get maximumWidth => this._maximumWidth;
  double _maximumWidth = double.nan;

  set maximumWidth(double newValue) => this._maximumWidth = newValue;

  double get width => this._width;
  double _width = double.nan;

  set width(double newValue) => this._width = newValue;

  EdgeInsetsGeometry get padding => this._padding;
  EdgeInsetsGeometry _padding = EdgeInsets.all(16);

  set padding(EdgeInsetsGeometry newValue) => this._padding = newValue;

  DataGridHeaderCellStyle get headerStyle => this._headerStyle;
  DataGridHeaderCellStyle _headerStyle;

  set headerStyle(DataGridHeaderCellStyle newHeaderStyle) =>
      this._headerStyle = newHeaderStyle;

  DataGridCellStyle get cellStyle => this._cellStyle;
  DataGridCellStyle _cellStyle;

  set cellStyle(DataGridCellStyle newCellStyle) =>
      this._cellStyle = newCellStyle;

  bool get isHidden => this._isHidden;
  bool _isHidden = false;

  set isHidden(bool newValue) => this._isHidden = newValue;

  String getFormattedValue(dynamic cellValue) {
    return cellValue?.toString();
  }

  @protected
  void setCellType(String cellType) {
    this._cellType = cellType;
  }
}

/// A column which displays the values of the string in its cells.
///
/// This column has all the required APIs to customize the widget [Text] as it displays [Text] for all the cells.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource,
///     columns: [
///     GridTextColumn(mappingName: 'name')
///     ..headerText = 'Name',
///     GridTextColumn(mappingName: 'designation')
///     ..headerText = 'Designation',
///   ],
///   );
/// }
/// ```
class GridTextColumn extends GridColumn {
  GridTextColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'TextField';
    this.headerTextAlignment = Alignment.centerLeft;
    this.overflow = TextOverflow.ellipsis;
  }
}

/// A column used to display the numeric values in its cells (int, double).
///
/// [GridNumericColumn] supports [numberFormat] for formatting numeric values.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     columns: [
///      GridNumericColumn(mappingName: 'id')
///      ..headerText = 'ID',
///      GridNumericColumn(mappingName: 'salary')
///      ..headerText = 'Salary',
///   ],
///   );
/// }
/// ```
class GridNumericColumn extends GridColumn {
  intl.NumberFormat get numberFormat => this._numberFormat;
  intl.NumberFormat _numberFormat;

  set numberFormat(intl.NumberFormat newNumberFormat) =>
      this._numberFormat = newNumberFormat;
  @override
  String getFormattedValue(dynamic cellValue) {
    if (this.numberFormat != null) {
      return this.numberFormat.format(cellValue);
    }
    return cellValue.toString();
  }

  GridNumericColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'Numeric';
    this.textAlignment = Alignment.centerRight;
    this.headerTextAlignment = Alignment.centerRight;
  }
}

/// A column that shows any custom widget inside its cells.
///
/// [GridWidgetColumn] supports loading [Image], [CheckBox] and so on in cells in column.
///
/// [SfDataGrid.cellBuilder] allows you to set the widget in a column as needed.
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     cellBuilder: (BuildContext context, GridColumn column, int rowIndex){
///        return employees[rowIndex].image;
///      },
///     columns: [
///      GridWidgetColumn(mappingName: image)
///      ..headerText = 'Image',
///   ],
///   );
/// }
/// ```
class GridWidgetColumn extends GridColumn {
  GridWidgetColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'Widget';
  }
}

/// A column which displays the values of DateTime in its cells.
///
/// [GridDateTimeColumn] supports [dateFormat] for formatting DateTime values.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     columns: [
///      GridNumericColumn(mappingName: 'id')
///      ..headerText = 'ID',
///      GridDateTimeColumn(mappingName: 'dateofjoining')
///      ..headerText = 'Date of Joining',
///   ],
///   );
/// }
/// ```
class GridDateTimeColumn extends GridColumn {
  intl.DateFormat get dateFormat => this._dateFormat;
  intl.DateFormat _dateFormat;

  set dateFormat(intl.DateFormat newDateTime) => this._dateFormat = newDateTime;

  final intl.DateFormat _defaultDateFormat = intl.DateFormat('dd-MM-yyyy');

  @override
  String getFormattedValue(dynamic cellValue) {
    if (this.dateFormat != null) {
      return this.dateFormat.format(cellValue);
    }
    return this._defaultDateFormat.format(cellValue);
  }

  GridDateTimeColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'DateTime';
    this.textAlignment = Alignment.centerRight;
    this.headerTextAlignment = Alignment.centerRight;
  }
}
