part of datagrid;

/// Used by Tree Table to balance the tree with algorithm based on Red - Black tree.
enum _TreeTableNodeColor {
  /// TreeTableNodeColoe.red, will represented the Red color.
  red,

  /// TreeTableNodeColoe.black, will represented the Black color.
  black
}

/// A branch or leaf in the tree.
abstract class _TreeTableNodeBase {
  /// Gets the parent branch.
  _TreeTableBranchBase get parent => _parent;
  _TreeTableBranchBase _parent;

  /// Gets the color to the branch.
  _TreeTableNodeColor get color => _color;
  _TreeTableNodeColor _color;

  /// Sets the color to the branch.
  set color(Object value) {
    if (value == _color) return;
    _color = value;
  }

  /// Sets the parent color.
  set parent(_TreeTableBranchBase value) {
    if (value == _parent) return;
    _parent = value;
  }

  /// Gets the position in the tree.
  ///
  /// Returns the position in the tree.
  int getPosition();

  ///Gets the number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  int getCount();

  /// Indicates whether leaf is empty.
  ///
  /// Returns the boolean value indicates whether the leaf is empty
  bool isEmpty();

  /// Indicates whether this is a leaf.
  ///
  /// Returns the boolean value indicates whether this is a leaf
  bool isEntry();

  /// Gets the tree level of this node.
  ///
  /// Returns  Returns the tree level of this node.
  int getLevel();

  /// Gets the minimum value (of the leftmost leaf) of the branch in a sorted tree.
  ///
  /// Returns the minimum value (of the leftmost leaf) of the branch in a sorted tree.
  Object getMinimum();

  /// Walk up parent branches and reset counters.
  ///
  /// * notifyParentRecordSource - _required_ - boolean value
  void invalidateCounterBottomUp(bool notifyParentRecordSource);

  /// Walk up parent branches and reset summaries.
  ///
  /// * notifyParentRecordSource - _required_ - Boolean value
  void invalidateSummariesBottomUp(bool notifyParentRecordSource);
}

/// A branch with left and right leaves or branches.
mixin _TreeTableBranchBase on _TreeTableNodeBase {
  /// Gets the left node.
  _TreeTableNodeBase get left => _left;
  _TreeTableNodeBase _left;

  /// Sets the left node.
  set left(_TreeTableNodeBase value) {
    if (value == _left) return;
    _left = value;
  }

  /// Sets the left node.
  ///
  /// Call this method instead of simply setting `Left` property if you want
  /// to avoid the round-trip call to check whether the tree is in add-mode
  /// or if tree-table is sorted.
  ///
  /// * value - _required_ - The new node.
  /// * inAddMode - _required_ - Indicates whether tree-table is in add-mode.
  /// * isSortedTree - _required_ - Indicates whether tree-table is sorted.
  void setLeft(_TreeTableNodeBase value, bool inAddMode, bool isSortedTree);

  /// Gets the right node.
  _TreeTableNodeBase get right => _right;
  _TreeTableNodeBase _right;

  /// Sets the right node.
  set right(_TreeTableNodeBase value) {
    if (value == _right) return;
    _right = value;
  }

  /// Sets the right node.
  ///
  /// Call this method instead of simply setting `Right` property if you want
  /// to avoid the round-trip call to check whether the tree is in add-mode
  /// or if tree-table is sorted.
  ///
  /// * value - _required_ - The new node.
  /// * inAddMode - _required_ - Specifies if tree-table is in add-mode.
  void setRight(_TreeTableNodeBase value, bool inAddMode);

  /// The left branch cast to _TreeTableBranchBase.
  ///
  /// Returns the left branch cast to _TreeTableBranchBase.
  _TreeTableBranchBase getLeftBranch();

  /// The right branch cast to _TreeTableBranchBase.
  ///
  /// Returns the right branch cast to _TreeTableBranchBase.
  _TreeTableBranchBase getRightBranch();

  /// Gets the Red-Black tree color.
  _TreeTableNodeColor get color => _color;
  _TreeTableNodeColor _color;

  /// Sets the Red-Balck tree color.
  set color(Object value) {
    if (value == _color) return;
    _color = value;
  }

  /// Returns the position in the tree table of the specified child node.
  ///
  /// * node - _required_ - Tree node
  ///
  /// Returns the position in the tree.
  int getEntryPositionOfChild(_TreeTableNodeBase node);

  /// Sets this object's child node Count dirty and
  /// marks parent nodes' child node Count dirty.
  void invalidateCountBottomUp();

  /// Sets this object's child node Count dirty and steps
  /// through all child branches and marks their child node Count dirty.
  void invalidateCountTopDown();

  /// Sets this object's child node Minimum dirty and
  /// marks parent nodes' child node Minimum dirty.
  void invalidateMinimumBottomUp();

  /// Sets this object's child node Minimum dirty and steps
  /// through all child branches and marks their child node Minimum dirty.
  void invalidateMinimumTopDown();
}

/// A leaf with value and optional sort key.
mixin _TreeTableEntryBase on _TreeTableNodeBase {
  /// Gets the sort key of this leaf.
  ///
  /// Returns the sort key of this leaf.
  Object getSortKey();

  /// Gets the value attached to this leaf.
  Object get value => _value;
  Object _value;

  /// Sets the value attached to this leaf
  set value(Object value) {
    if (value == _value) return;
    _value = value;
  }

  /// Creates a branch that can hold this entry when new leaves are inserted into the tree.
  ///
  /// * tree - _required_ - Tree table instance
  ///
  /// Returns the instance of newly created branch
  _TreeTableBranchBase createBranch(_TreeTable tree);
}

/// A branch or leaf in the tree.
abstract class _TreeTableNode extends _TreeTableNodeBase {
  _TreeTableBranchBase _parent;
  static Object emptyMin = Object();

  /// Gets the tree this node belongs to.
  _TreeTable get tree => _tree;
  _TreeTable _tree;

  /// Sets the tree this node belongs to.
  set tree(_TreeTable value) {
    if (value == _tree) return;
    _tree = value;
  }

  /// Gets the parent branch.
  _TreeTableBranchBase get parent => this._parent;

  /// Sets the parent branch.
  set parent(_TreeTableBranchBase value) {
    this._parent = value;
  }

  /// Gets the position in the tree.
  ///
  /// Returns the position in the tree.
  @override
  int getPosition() {
    if (this.parent == null) {
      return 0;
    }

    return this.parent.getEntryPositionOfChild(this);
  }

  /// Gets the minimum value (of the most-left leaf) of the branch in a sorted tree.
  ///
  /// Returns the minimum value (of the most-left leaf) of the branch in a sorted tree.
  @override
  Object getMinimum() {
    return emptyMin;
  }

  /// Indicates whether leaf is empty.
  ///
  /// Returns the boolean value indicates whether the leaf is empty
  @override
  bool isEmpty() {
    return this.getCount() == 0;
  }

  /// Indicates whether this is a leaf.
  ///
  /// Returns the boolean value indicates whether this is a leaf
  bool isEntry();

  /// Gets the number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  int getCount();

  /// Gets the tree level of this node.
  ///
  /// Returns the tree level of this node.
  int getLevel() {
    int level = 0;
    if (this.parent != null) {
      level = this.parent.getLevel() + 1;
    }

    return level;
  }

  /// Gets the Debug / text information about the node.
  ///
  /// Returns the Debug / text information about the node.
  String getNodeInfoBase() {
    String side = "_";
    if (this.parent != null) {
      side = _MathHelper.referenceEquals(this.parent.left, this) ? "L" : "R";
    }

    return this.getLevel().toString() +
        "," +
        side +
        "this.," +
        this.getPosition().toString() +
        ", " +
        this.getCount().toString();
  }

  /// Gets the Debug / text information about the node.
  ///
  /// Returns the Debug / text information about the node.
  @override
  String toString() {
    return '$runtimeType{${this.getNodeInfoBase()} + }';
  }

  /// Walks up parent branches and reset counters.
  ///
  /// * notifyParentRecordSource - _required_ - Boolean value
  @override
  void invalidateCounterBottomUp(bool notifyParentRecordSource) {}

  /// Walks up parent branches and reset summaries.
  ///
  /// * notifyParentRecordSource - _required_ - Boolean value
  @override
  void invalidateSummariesBottomUp(bool notifyParentRecordSource) {}
}

/// A branch in a tree.
class _TreeTableBranch extends _TreeTableNode with _TreeTableBranchBase {
  _TreeTableNodeBase _left;
  _TreeTableNodeBase _right;
  int entryCount = -1;
  Object _minimum = _TreeTableNode.emptyMin;

  /// Initializes a new instance of the `TreeTableBranch` class.
  ///
  /// * tree - _required_ - Tree table instance
  _TreeTableBranch(_TreeTable tree) {
    this.tree = tree;
  }

  /// Gets the minimum value (of the most-left leaf) of the branch in a sorted tree.
  ///
  /// Returns the minimum value (of the most-left leaf) of the branch in a sorted tree.
  @override
  Object getMinimum() {
    if (_MathHelper.referenceEquals(_TreeTableNode.emptyMin, this._minimum)) {
      this._minimum = this._left.getMinimum();
    }
    return this._minimum;
  }

  /// Gets Red-Black tree algorithm helper.
  _TreeTableNodeColor get color => _color;
  _TreeTableNodeColor _color;

  /// Sets Red-black tree algorithm helper.
  set color(Object value) {
    if (value == _color) return;
    _color = value;
  }

  /// The left node cast to _TreeTableBranchBase.
  ///
  /// Returns the left node cast to _TreeTableBranchBase.
  _TreeTableBranchBase getLeftBranch() {
    return this._left as _TreeTableBranchBase;
  }

  /// The right node cast to _TreeTableBranchBase.
  ///
  /// Returns the right node cast to _TreeTableBranchBase.
  _TreeTableBranchBase getRightBranch() {
    return this._right as _TreeTableBranchBase;
  }

  /// Gets the left leaf or branch.
  _TreeTableNodeBase get left => this._left;

  /// Sets the left leaf or branch.
  set left(_TreeTableNodeBase value) {
    this.setLeft(value, false, tree.sorted);
  }

  /// Sets the left node.
  ///
  /// Call this method instead of simply setting `Left` property if you want
  /// to avoid the round-trip call to check whether the tree is in add-mode
  /// or if tree-table is sorted.
  ///
  /// * value - _required_ - The new node.
  /// * inAddMode - _required_ - Indicates whether tree-table is in add-mode.
  /// * isSorted - _required_ - Indicates whether tree-table is sorted.
  @override
  void setLeft(_TreeTableNodeBase value, bool inAddMode, bool isSorted) {
    if (!_MathHelper.referenceEquals(this.left, value)) {
      if (inAddMode) {
        if (this._left != null && this._left.parent == this) {
          this._left.parent = null;
        }

        this._left = value;
        if (this._left != null) {
          this._left.parent = this;
        }
      } else {
        int lc = (this._left != null) ? this._left.getCount() : 0;
        int vc = (value != null) ? value.getCount() : 0;
        int entryCountDelta = vc - lc;
        if (this._left != null && this._left.parent == this) {
          this._left.parent = null;
        }

        this._left = value;
        if (this._left != null) {
          this._left.parent = this;
        }

        if (entryCountDelta != 0) {
          this.invalidateCountBottomUp();
        }

        if (isSorted) {
          this.invalidateMinimumBottomUp();
        }

        this.invalidateCounterBottomUp(false);
        this.invalidateSummariesBottomUp(false);
      }
    }
  }

  /// Gets the right tree or branch.
  _TreeTableNodeBase get right => this._right;

  /// Sets the right tree or branch.
  set right(_TreeTableNodeBase value) {
    this.setRight(value, false);
  }

  /// Sets the right node.
  ///
  /// Call this method instead of simply setting `Right` property if you want
  /// to avoid the round-trip call to check whether the tree is in add-mode
  /// or if tree-table is sorted.
  ///
  /// * value - _required_ - The new node.
  /// * inAddMode - _required_ - Indicates whether tree-table is in add-mode.
  @override
  void setRight(_TreeTableNodeBase value, bool inAddMode) {
    if (!_MathHelper.referenceEquals(this.right, value)) {
      if (inAddMode) {
        if (this._right != null && this._right.parent == this) {
          this._right.parent = null;
        }

        this._right = value;
        if (this._right != null) {
          this._right.parent = this;
        }
      } else {
        int lc = (this._right != null) ? this._right.getCount() : 0;
        int vc = (value != null) ? value.getCount() : 0;
        int entryCountDelta = vc - lc;
        if (this._right != null && this._right.parent == this) {
          this._right.parent = null;
        }

        this._right = value;
        if (this._right != null) {
          this._right.parent = this;
        }

        if (entryCountDelta != 0) {
          this.invalidateCountBottomUp();
        }

        this.invalidateCounterBottomUp(false);
        this.invalidateSummariesBottomUp(false);
      }
    }
  }

  /// Gets the position in the tree table of the specific child node.
  ///
  /// * node - _required_ - Tree node
  ///
  /// Returns the position in the tree table of the specific child node.
  @override
  int getEntryPositionOfChild(_TreeTableNodeBase node) {
    int pos = getPosition();
    if (_MathHelper.referenceEquals(node, this.right)) {
      pos += this.left.getCount();
    } else if (!_MathHelper.referenceEquals(node, this.left)) {
      //throw ArgumentError('must be a child node','node');
      throw ArgumentError("must be a child node");
    }

    return pos;
  }

  /// Gets the number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  @override
  int getCount() {
    if (this.entryCount < 0) {
      this.entryCount = this._left.getCount() + this._right.getCount();
    }

    return this.entryCount;
  }

  /// Indicates whether this is a leaf.
  ///
  /// Returns boolean value
  @override
  bool isEntry() {
    return false;
  }

  /// Sets this object's child node count dirty and
  /// walks up parent nodes and marks their child node count dirty.
  @override
  void invalidateCountBottomUp() {
    this.entryCount = -1;
    if (this.parent != null && parent.parent == this.parent) {
      throw Exception();
    }

    if (this.parent != null) {
      parent.invalidateCountBottomUp();
    }
  }

  /// Sets this object's child node count dirty and steps
  /// through all child branches and marks their child node count dirty.
  @override
  void invalidateCountTopDown() {
    this.entryCount = -1;
    if (!this.left.isEntry()) {
      this.getLeftBranch().invalidateCountTopDown();
    }

    if (!this.right.isEntry()) {
      this.getRightBranch().invalidateCountTopDown();
    }
  }

  /// Sets this object's child node minimum dirty and
  /// marks parent nodes' child node minimum dirty.
  @override
  void invalidateMinimumBottomUp() {
    this._minimum = _TreeTableNode.emptyMin;
    if (this.parent != null) {
      parent.invalidateMinimumBottomUp();
    }
  }

  /// Sets this object's child node minimum dirty and steps
  /// through all child branches and marks their child node minimum dirty.
  @override
  void invalidateMinimumTopDown() {
    if (!this.left.isEntry()) {
      this.getLeftBranch().invalidateMinimumTopDown();
    }
    if (!this.right.isEntry()) {
      this.getRightBranch().invalidateMinimumTopDown();
    }
    this._minimum = _TreeTableNode.emptyMin;
  }
}

/// A leaf in the tree with value and optional sort key.
class _TreeTableEntry extends _TreeTableNode with _TreeTableEntryBase {
  /// Gets the value attached to this leaf.
  @override
  Object _value;
  Object get value => this._value;

  /// Sets the value attached to this leaf.
  set value(Object value) {
    this._value = value;
  }

  /// Gets the sort key of this leaf.
  ///
  /// Returns the sort key of this leaf.
  Object getSortKey() {
    return this.value;
  }

  /// Gets the minimum value (of the most-left leaf) of the branch in a sorted tree.
  ///
  /// Returns the minimum value (of the most-left leaf) of the branch in a sorted tree.
  @override
  Object getMinimum() {
    return this.getSortKey();
  }

  /// Gets the number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  @override
  int getCount() {
    return 1;
  }

  /// Creates a branch that can hold this entry when new leaves are inserted into the tree.
  ///
  /// * tree - _required_ - Tree table instance
  ///
  /// Returns the instance of newly created branch

  @override
  _TreeTableBranchBase createBranch(_TreeTable tree) {
    return _TreeTableBranch(tree);
  }

  /// Gets the Debug / text information about the node.
  ///
  /// Returns the Debug / text information about the node.
  String getNodeInfo() {
    return getNodeInfoBase() +
        ", " +
        (this.value != null ? this.value.toString() : 'null');
  }

  /// Indicates whether this is a leaf.
  ///
  /// Returns the boolean value
  @override
  bool isEntry() {
    return true;
  }
}

/// An empty node.
class _TreeTableEmpty extends _TreeTableNode {
  static _TreeTableEmpty empty = _TreeTableEmpty();

  /// The number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  @override
  int getCount() {
    return 0;
  }

  /// Indicates whether this is a leaf.
  ///
  /// Returns boolean value
  @override
  bool isEntry() {
    return true;
  }

  /// The Debug / text information about the node.
  ///
  /// Returns the Debug / text information about the node.
  String getNodeInfo() {
    return "Empty";
  }
}

/// Tree table interface definition.
class _TreeTableBase extends _ListBase {
  /// Gets a value indicating whether this is a sorted tree or not.
  bool get sorted => _sorted;
  bool _sorted = false;

  /// Gets the comparer value used by sorted trees.
  Comparable get comparer => _comparer;
  Comparable _comparer;

  /// Sets the comparer value used by sorted trees.
  set comparer(Comparable value) {
    if (value == _comparer) return;
    _comparer = value;
  }

  /// Gets the root node.
  _TreeTableNodeBase get root => _root;
  _TreeTableNodeBase _root;

  /// Gets a value indicating whether the tree was initialize or not.
  bool get isInitializing => _isInitializing;
  bool _isInitializing = false;

  /// Optimizes insertion of many elements when tree is initialized for the first time.
  void beginInit();

  /// Ends optimization of insertion of elements when tree is initialized for the first time.
  void endInit();

  /// Optimized access to a subsequent entry.
  ///
  /// * current - _required_ - Current item
  ///
  /// Returns next subsequent entry
  _TreeTableEntryBase getNextEntry(_TreeTableEntryBase current);

  /// Optimized access to a previous entry.
  ///
  /// * current - _required_ - Current item
  ///
  /// Returns previous entry
  _TreeTableEntryBase getPreviousEntry(_TreeTableEntryBase current);
}

/// A tree table.
class _TreeTable extends _TreeTableBase {
  _TreeTableNodeBase _root;
  Object _tag;
  Comparable _comparer;
  bool _sorted = false;
  bool _inAddMode = false;
  _TreeTableBranchBase _lastAddBranch;
  _TreeTableEntryBase _lastIndexLeaf;

  /// Gets the last index leaf.
  ///
  /// Returns the last index leaf.
  _TreeTableEntryBase get lastIndexLeaf => this._lastIndexLeaf;

  /// Sets the last index leaf.
  ///
  /// Returns the last index leaf.
  set lastIndexLeaf(_TreeTableEntryBase value) {
    if (this._lastIndexLeaf != value) this._lastIndexLeaf = value;
  }

  void lastIndexLeafDisposed() {
    this.lastIndexLeaf = null;
    this.lastIndex = -1;
  }

  int lastIndex = -1;

  /// Gets the comparer used by sorted trees.
  @override
  Comparable get comparer => this._comparer;

  /// Sets the comparer used by sorted trees.
  set comparer(Comparable value) {
    this._comparer = value;
    this._sorted = this._comparer != null;
  }

  /// Gets the tag that can be associated with this object.
  Object get tag => this._tag;

  /// Sets the tag that can be associated with this object.
  set tag(Object value) {
    this._tag = value;
  }

  /// Initializes a new instance of the `TreeTable` class.
  ///
  /// * sorted - _required_ - Boolean value
  _TreeTable(bool sorted) {
    this._sorted = sorted;
    this._inAddMode = false;
  }

  /// Gets a value indicating whether tree is sorted or not.
  bool get sorted => this._sorted;

  /// Gets the root node.
  _TreeTableNodeBase get root => _root;

  /// Seta the root node.
  set root(_TreeTableNodeBase value) {
    this._root = value;
  }

  void leftRotate(_TreeTableBranchBase x, bool inAddMode) {
    var y = x.right as _TreeTableBranchBase;
    if (y == null) {
      return;
    }
    _TreeTableNodeBase yLeft = y.left;
    y.setLeft(_TreeTableEmpty.empty, inAddMode, this.sorted);
    x.setRight(yLeft, inAddMode);
    if (x.parent != null) {
      if (_MathHelper.referenceEquals(x, x.parent.left)) {
        x.parent.setLeft(y, inAddMode, this.sorted);
      } else {
        x.parent.setRight(y, inAddMode);
      }
    } else {
      this._root = y;
    }
    y.setLeft(x, inAddMode, this.sorted);
  }

  void rightRotate(_TreeTableBranchBase x, bool inAddMode) {
    var y = x.left as _TreeTableBranchBase;
    if (y == null) {
      return;
    }
    _TreeTableNodeBase yRight = y.right;
    y.setRight(_TreeTableEmpty.empty,
        inAddMode); // make sure Parent is not reset later
    x.setLeft(yRight, inAddMode, this.sorted);
    if (x.parent != null) {
      if (x == x.parent.right) {
        x.parent.setRight(y, inAddMode);
      } else {
        x.parent.setLeft(y, inAddMode, this.sorted);
      }
    } else {
      this.root = y;
    }
    y.setRight(x, inAddMode);
  }

  void insertFixup(_TreeTableBranchBase x, bool inAddMode) {
    // Check Red-Black properties
    while (!_MathHelper.referenceEquals(x, this._root) &&
        x.parent.color == _TreeTableNodeColor.red &&
        x.parent.parent != null) {
      // We have a violation
      if (x.parent == x.parent.parent.left) {
        var y = x.parent.parent.right as _TreeTableBranchBase;
        if (y != null && y.color == _TreeTableNodeColor.red) {
          // uncle is red
          x.parent.color = _TreeTableNodeColor.black;
          y.color = _TreeTableNodeColor.black;
          x.parent.parent.color = _TreeTableNodeColor.red;
          x = x.parent.parent;
        } else {
          // uncle is black
          if (x == x.parent.right) {
            // Make x a left child
            x = x.parent;
            this.leftRotate(x, inAddMode);
          }

          // Recolor and rotate
          x.parent.color = _TreeTableNodeColor.black;
          x.parent.parent.color = _TreeTableNodeColor.red;
          this.rightRotate(x.parent.parent, inAddMode);
        }
      } else {
        // Mirror image of above code
        var y = x.parent.parent.left as _TreeTableBranchBase;
        if (y != null && y.color == _TreeTableNodeColor.red) {
          // uncle is red
          x.parent.color = _TreeTableNodeColor.black;
          y.color = _TreeTableNodeColor.black;
          x.parent.parent.color = _TreeTableNodeColor.red;
          x = x.parent.parent;
        } else {
          // uncle is black
          if (x == x.parent.left) {
            x = x.parent;
            this.rightRotate(x, inAddMode);
          }

          x.parent.color = _TreeTableNodeColor.black;
          x.parent.parent.color = _TreeTableNodeColor.red;
          this.leftRotate(x.parent.parent, inAddMode);
        }
      }
    }

    (this.root).color = _TreeTableNodeColor.black;
  }

  void deleteFixup(_TreeTableBranchBase x, bool isLeft) {
    bool inAddMode = false;
    while (!_MathHelper.referenceEquals(x, this._root) &&
        x._color == _TreeTableNodeColor.black) {
      if (isLeft) {
        var w = x.parent.right as _TreeTableBranchBase;
        if (w != null && w.color == _TreeTableNodeColor.red) {
          w.color = _TreeTableNodeColor.black;
          x.parent.color = _TreeTableNodeColor.black;
          this.leftRotate(x.parent, inAddMode);
          w = x.parent.right as _TreeTableBranchBase;
        }

        if (w == null) {
          return;
        }

        if (w.color == _TreeTableNodeColor.black &&
            (w.left.isEntry() ||
                w.getLeftBranch().color == _TreeTableNodeColor.black) &&
            (w.right.isEntry() ||
                w.getRightBranch().color == _TreeTableNodeColor.black)) {
          w.color = _TreeTableNodeColor.red;
          if (x.color == _TreeTableNodeColor.red) {
            x.color = _TreeTableNodeColor.black;
            return;
          } else {
            isLeft = x.parent.left == x;
            x = x.parent;
          }
        } else if (w.color == _TreeTableNodeColor.black &&
            !w.right.isEntry() &&
            w.getRightBranch().color == _TreeTableNodeColor.red) {
          this.leftRotate(x.parent, inAddMode);
          _TreeTableNodeColor t = w.color;
          w.color = x.parent.color;
          x.parent.color = t;
          return;
        } else if (w.color == _TreeTableNodeColor.black &&
            !w.left.isEntry() &&
            w.getLeftBranch().color == _TreeTableNodeColor.red &&
            (w.right.isEntry() ||
                w.getRightBranch().color == _TreeTableNodeColor.black)) {
          this.rightRotate(w, inAddMode);

          w.parent.color = _TreeTableNodeColor.black;
          w.color = _TreeTableNodeColor.red;

          this.leftRotate(x.parent, inAddMode);
          _TreeTableNodeColor t = w.color;
          w.color = x.parent.color;
          x.parent.color = t;
          return;
        } else {
          return;
        }
      } else {
        var w = x.parent.left as _TreeTableBranchBase;
        if (w != null && w.color == _TreeTableNodeColor.red) {
          w.color = _TreeTableNodeColor.black;
          x.parent.color = _TreeTableNodeColor.red;
          this.rightRotate(x.parent, inAddMode);
          w = x.parent.left as _TreeTableBranchBase;
        }

        if (w == null) {
          return;
        }

        if (w.color == _TreeTableNodeColor.black &&
            (w.left.isEntry() ||
                w.getLeftBranch().color == _TreeTableNodeColor.black) &&
            (w.right.isEntry() ||
                w.getRightBranch().color == _TreeTableNodeColor.black)) {
          w.color = _TreeTableNodeColor.red;
          if (x.color == _TreeTableNodeColor.red) {
            x.color = _TreeTableNodeColor.black;
            return;
          } else if (x.parent != null) {
            isLeft = x.parent.left == x;
            x = x.parent;
          }
        } else {
          if (w.color == _TreeTableNodeColor.black &&
              !w.right.isEntry() &&
              w.getRightBranch().color == _TreeTableNodeColor.red) {
            _TreeTableBranchBase xParent = x.parent;
            this.leftRotate(xParent, inAddMode);
            _TreeTableNodeColor t = w.color;
            w.color = xParent.color;
            xParent.color = t;
            return;
          } else if (w.color == _TreeTableNodeColor.black &&
              !w.left.isEntry() &&
              w.getLeftBranch().color == _TreeTableNodeColor.red &&
              (w.right.isEntry() ||
                  w.getRightBranch().color == _TreeTableNodeColor.black)) {
            _TreeTableBranchBase wParent = w.parent;
            _TreeTableBranchBase xParent = x.parent;
            this.rightRotate(w, inAddMode);

            wParent.color = _TreeTableNodeColor.black;
            w.color = _TreeTableNodeColor.red;

            this.leftRotate(x.parent, inAddMode);
            _TreeTableNodeColor t = w.color;
            w.color = xParent.color;
            xParent.color = t;
            return;
          }
        }
      }
    }

    x.color = _TreeTableNodeColor.black;
  }

  /// Gets an item at the specified index.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns the item at the specified index.
  operator [](index) {
    this._getEntryAt(index);
  }

  /// Sets an item at the specified index.
  operator []=(index, value) {
    this.setNodeAt(index, value);
  }

  /// Sets the node at the specified index.
  ///
  /// * index - _required_ - Index value where the node is to be inserted.
  /// * value - _required_ - Value of the node that is to be inserted.
  void setNodeAt(int index, _TreeTableNodeBase value) {
    _TreeTableEntryBase leaf = this._getEntryAt(index);
    if (_MathHelper.referenceEquals(leaf, this._root)) {
      this._root = value;
    } else {
      _TreeTableBranchBase branch = leaf.parent;
      this._replaceNode(branch, leaf, value, false);
    }

    this.lastIndex = -1;
  }

  _TreeTableEntryBase _getEntryAt(int index) {
    int treeCount = this.getCount();
    if (index < 0 || index >= treeCount) {
      throw new ArgumentError(
          'index ${index.toString()}  must be between 0 and ${(treeCount - 1).toString()}');
    }

    if (this._root == null) {
      // replace root
      return this._root as _TreeTableEntryBase;
    } else {
      if (this.lastIndex != -1) {
        if (index == this.lastIndex) {
          return this._lastIndexLeaf;
        } else if (index == this.lastIndex + 1) {
          this.lastIndex++;
          this._lastIndexLeaf = this.getNextEntry(this._lastIndexLeaf);
          return this._lastIndexLeaf;
        }
      }

      // find node
      _TreeTableBranchBase branch;
      _TreeTableNodeBase current = this._root;
      int count = 0;
      while (!current.isEntry()) {
        branch = current as _TreeTableBranchBase;
        int leftCount = branch.left.getCount();

        if (index < count + leftCount) {
          current = branch.left;
        } else {
          count += branch.left.getCount();
          current = branch.right;
        }
      }
      this.lastIndexLeaf = current as _TreeTableEntryBase;
      this.lastIndex = index;
      return this._lastIndexLeaf;
    }
  }

  /// Optimized access to the previous entry.
  ///
  /// * current - _required_ - Current item
  ///
  /// Returns previous entry
  _TreeTableEntryBase getPreviousEntry(_TreeTableEntryBase current) {
    _TreeTableBranchBase parent = current.parent;
    _TreeTableNodeBase prev;

    if (parent == null) {
      prev = null;
      return null;
    } else {
      if (_MathHelper.referenceEquals(current, parent.right)) {
        prev = parent.left;
      } else {
        _TreeTableBranchBase parentParent = parent.parent;
        if (parentParent == null) {
          return null;
        } else {
          while (_MathHelper.referenceEquals(parentParent.left, parent)) {
            parent = parentParent;
            parentParent = parentParent.parent;
            if (parentParent == null) {
              return null;
            }
          }

          prev = parentParent.left;
        }
      }

      while (!prev.isEntry()) {
        prev = (prev as _TreeTableBranchBase).right;
      }
    }

    return prev as _TreeTableEntryBase;
  }

  /// Optimized access to a subsequent entry.
  ///
  /// * current - _required_ - Current item
  ///
  /// Returns next subsequent entry
  _TreeTableEntryBase getNextEntry(_TreeTableEntryBase current) {
    _TreeTableBranchBase parent = current.parent;
    _TreeTableNodeBase next;

    if (parent == null) {
      next = null;
      return null;
    } else {
      if (_MathHelper.referenceEquals(current, parent.left)) {
        next = parent.right;
      } else {
        _TreeTableBranchBase parentParent = parent.parent;
        if (parentParent == null) {
          return null;
        } else {
          while (_MathHelper.referenceEquals(parentParent.right, parent)) {
            parent = parentParent;
            parentParent = parentParent.parent;
            if (parentParent == null) {
              return null;
            }
          }

          next = parentParent.right;
        }
      }

      while (!next.isEntry()) {
        next = (next as _TreeTableBranchBase).left;
      }
    }

    return next as _TreeTableEntryBase;
  }

  _TreeTableEntryBase getMostLeftEntry(_TreeTableBranchBase parent) {
    _TreeTableNodeBase next;

    if (parent == null) {
      next = null;
      return null;
    } else {
      next = parent.left;
      while (!next.isEntry()) {
        next = (next as _TreeTableBranchBase).left;
      }
    }

    return next as _TreeTableEntryBase;
  }

  /// Inserts a node at the specified index.
  ///
  /// * index - _required_ - Index value where the node is to be inserted.
  /// * value - _required_ - Value of the node to insert.
  void insertBase(int index, _TreeTableNodeBase value) {
    if (this.sorted) {
      throw Exception("This tree is sorted - use AddSorted instead.");
    }

    int treeCount = this.getCount();
    if (index < 0 || index > treeCount) {
      throw new ArgumentError(
          'index ${index.toString()} must be between 0 and ${treeCount.toString()}');
    }

    if (index == treeCount) {
      this.add(value);
      return;
    }

    this.cacheLastFoundEntry(null, null, false);
    if (this._root == null) {
      // replace root
      this._root = value;
    } else {
      _TreeTableEntryBase leaf;
      if (this.lastIndex != -1) {
        if (index == this.lastIndex) {
          leaf = this.lastIndexLeaf;
        } else if (index == this.lastIndex + 1) {
          leaf = this.getNextEntry(this.lastIndexLeaf);
        }
      }
      if (leaf == null) {
        leaf = this._getEntryAt(index);
      }

      _TreeTableBranchBase branch = leaf.parent;
      _TreeTableBranchBase newBranch = leaf.createBranch(this);

      newBranch.setLeft(value, false, this.sorted); // will set leaf.Parent ...
      newBranch.right = leaf;

      if (branch == null) {
        this.root = newBranch;
      } else {
        // swap out leafs parent with new node
        this._replaceNode(branch, leaf, newBranch, false);
      }

      this.insertFixup(newBranch, this._inAddMode);

      if (value.isEntry()) {
        this._lastIndexLeaf = value as _TreeTableEntryBase;
        this.lastIndex = index;
      } else {
        this._lastIndexLeaf = null;
        this.lastIndex = -1;
      }
    }
  }

  /// Removes the specified node.
  ///
  /// * value - _required_ - Node value to look for and remove.
  ///
  /// Returns the removed value
  bool removeBase(_TreeTableNodeBase value) {
    return this._remove(value, true);
  }

  /// Resets the cache.
  void resetCache() {
    this._lastAddBranch = null;
    this.lastIndex = -1;
    this._lastIndexLeaf = null;
  }

  /// Used to remove the value from the tree table
  ///
  /// * value - _required_ - Tree value
  /// * resetParent - _required_ - Boolean value
  ///
  /// Returns the boolean value
  bool _remove(_TreeTableNodeBase value, bool resetParent) {
    if (value == null) {
      return false;
    }

    if (!this.contains(value)) {
      return false;
    }

    this.cacheLastFoundEntry(null, null, false);

    this._lastAddBranch = null;
    this.lastIndex = -1;
    this._lastIndexLeaf = null;

    // root
    if (_MathHelper.referenceEquals(value, this.root)) {
      this._root = null;
      if (resetParent) {
        value.parent = null;
      }
    } else {
      _TreeTableBranchBase leafsParent = value.parent;

      // get the sister node
      _TreeTableNodeBase sisterNode = this._getSisterNode(leafsParent, value);

      // swap out leaves parent with sister
      if (_MathHelper.referenceEquals(leafsParent, this._root)) {
        this._root = sisterNode;
        this._root.parent = null;
      } else {
        _TreeTableBranchBase leafsParentParent = leafsParent.parent;
        bool isLeft = leafsParentParent.left == leafsParent;
        this._replaceNode(leafsParentParent, leafsParent, sisterNode, false);

        if (leafsParent.color == _TreeTableNodeColor.black) {
          leafsParent.parent = leafsParentParent;
          this.deleteFixup(leafsParent, isLeft);
        }
      }

      if (resetParent) {
        value.parent = null;
      }
    }

    return true;
  }

  _TreeTableNodeBase _getSisterNode(
      _TreeTableBranchBase leafsParent, _TreeTableNodeBase node) {
    _TreeTableNodeBase sisterNode;
    sisterNode = _MathHelper.referenceEquals(leafsParent.left, node)
        ? leafsParent.right
        : leafsParent.left;
    return sisterNode;
  }

  void _replaceNode(_TreeTableBranchBase branch, _TreeTableNodeBase oldNode,
      _TreeTableNodeBase newNode, bool inAddMode) {
    // also updates node count.
    if (_MathHelper.referenceEquals(branch.left, oldNode)) {
      branch.setLeft(newNode, inAddMode, this.sorted);
    } else {
      branch.setRight(newNode, inAddMode);
    }
  }

  /// Indicates whether the node belongs to this tree.
  ///
  /// * value - _required_ - Node value to search for.
  ///
  /// Returns true if node belongs to this tree; false otherwise.

  bool containsBase(_TreeTableNodeBase value) {
    if (value == null || this._root == null) {
      return false;
    }

    // search root
    while (value.parent != null) {
      value = value.parent;
    }

    return _MathHelper.referenceEquals(value, this._root);
  }

  /// Returns the position of a node.
  ///
  /// * value - _required_ - Node value to look for.
  ///
  /// Returns Index of the node if found.
  int indexOfBase(_TreeTableNodeBase value) {
    if (!this.contains(value)) {
      return -1;
    }

    return value.getPosition();
  }

  /// Appends a node.
  ///
  /// * value - _required_ - Node value to append.
  ///
  /// Returns the zero-based collection index at which the value has been added.
  int addBase(_TreeTableNodeBase value) {
    this.cacheLastFoundEntry(null, null, false);
    this.lastIndex = -1;

    if (this.sorted && !this._inAddMode) {
      return this.addSorted(value);
    }

    if (this._root == null) {
      // replace root
      this._root = value;
      return 0;
    } else {
      // add node to most right branch
      _TreeTableBranchBase branch;
      _TreeTableNodeBase current = this._lastAddBranch ?? this._root;

      while (!current.isEntry()) {
        branch = current as _TreeTableBranchBase;

        current = branch.right;
      }

      var leaf = current as _TreeTableEntryBase;

      _TreeTableBranchBase newBranch = leaf.createBranch(this);

      newBranch.setLeft(leaf, this._inAddMode, this.sorted);

      // will set leaf.Parent ...
      newBranch.setRight(value, this._inAddMode);

      if (branch == null) {
        this._root = newBranch;
      } else {
        // swap out leafs parent with new node
        this._replaceNode(branch, current, newBranch, this._inAddMode);
        if (!(branch.parent == null ||
            branch.parent.parent == null ||
            branch.right != branch.parent.parent.right)) {
          throw Exception();
        }

        if (!(branch.parent == null ||
            branch.parent.left.isEntry() ||
            (branch.parent.left as _TreeTableBranch).right != branch)) {
          throw Exception();
        }
      }

      this.insertFixup(newBranch, this._inAddMode);

      if (value.parent != null && value.parent.parent != null) {
        if (value.parent.parent.right == value) {
          throw new Exception();
        }
      }
      this._lastAddBranch = newBranch;

      if (this._inAddMode) {
        return -1;
      } else {
        return this._root.getCount() - 1;
      }
    }
  }

  /// Gets a value indicating whether the tree was initialize or not.
  bool get isInitializing => this._inAddMode;

  /// Optimizes insertion of many elements when tree is initialized for the first time.
  @override
  void beginInit() {
    this._inAddMode = true;
  }

  /// Ends optimization of insertion of elements when tree is initialized for the first time.
  @override
  void endInit() {
    this._inAddMode = false;

    // Fixes issues when GetCount() was called while debugging ...
    var branch = this._root as _TreeTableBranch;
    if (branch != null && branch.entryCount != -1) {
      branch.entryCount = -1;
    }
  }

  /// Adds a node into a sorted tree.
  ///
  /// * value - _required_ - Node value to add.
  ///
  /// Returns the zero-based collection index at which the value has been added.
  int addSorted(_TreeTableNodeBase value) {
    if (!this.sorted) {
      throw FormatException("This tree is not sorted.");
    }

    if (this._inAddMode) {
      return this.add(value);
    }

    this.cacheLastFoundEntry(null, null, false);

    if (this._root == null) {
      // replace root
      this._root = value;
      return 0;
    } else {
      bool inAddMode = false;
      Comparable comparer = this.comparer;

      // find node
      _TreeTableBranchBase branch;
      _TreeTableNodeBase current = this._root;
      int count = 0;
      int cmp = 0;

      while (!current.isEntry()) {
        branch = current as _TreeTableBranchBase;
        if (comparer != null) {
          cmp =
              Comparable.compare(value.getMinimum(), branch.right.getMinimum());
        } else if (value.getMinimum() is Comparable) {
          cmp = (value.getMinimum() as Comparable)
              .compareTo(branch.right.getMinimum());
        } else {
          throw Exception('No Comparer Specified');
        }

        if (cmp <= 0) {
          current = branch.left;
        } else {
          count += branch.left.getCount();
          current = branch.right;
        }
      }

      var leaf = current as _TreeTableEntryBase;

      _TreeTableBranchBase newBranch = leaf.createBranch(this);

      if (comparer != null) {
        cmp = Comparable.compare(value.getMinimum(), leaf.getSortKey());
      } else if (value.getMinimum() is Comparable) {
        cmp = (value.getMinimum() as Comparable).compareTo(leaf.getSortKey());
      }

      if (cmp <= 0) {
        newBranch.setLeft(
            value, false, this.sorted); // will set leaf.Parent ...
        newBranch.right = leaf;
      } else {
        newBranch.setLeft(leaf, false, this.sorted); // will set leaf.Parent ...
        count++;
        newBranch.right = value;
      }

      if (branch == null) {
        this._root = newBranch;
      } else {
        // swap out leafs parent with new node
        this._replaceNode(branch, leaf, newBranch, inAddMode);
      }

      //Debug.Assert(value.Position == index);
      this.insertFixup(newBranch, inAddMode);
      return count;
    }
  }

  /// Adds a node in a sorted tree only if no node with the same value has not been added yet.
  ///
  /// * key - _required_ - Key needs to be add in the collection
  /// * value - _required_ - Node value to add.
  ///
  /// Returns the instance for the tree
  _TreeTableEntryBase addIfNotExists(Object key, _TreeTableEntryBase value) {
    if (!this.sorted) {
      throw Exception("This tree is not sorted.");
    }

    this.cacheLastFoundEntry(null, null, false);

    if (this._root == null) {
      // replace root
      this._root = value;
      return value;
    } else {
      // find node
      _TreeTableBranchBase branch;
      _TreeTableNodeBase current = this._root;
      int cmp = 0;
      Comparable comparer = this.comparer;
      bool inAddMode = false;
      var comparableKey = key as Comparable;
      while (!current.isEntry()) {
        branch = current as _TreeTableBranchBase;
        if (comparer != null) {
          cmp = Comparable.compare(key, branch.right.getMinimum());
        } else if (comparableKey != null) {
          cmp = comparableKey.compareTo(branch.right.getMinimum());
        } else {
          throw Exception("No Comparer specified.");
        }

        if (cmp == 0) {
          current = branch.right;
          while (!current.isEntry()) {
            current = (current as _TreeTableBranchBase).left;
          }

          return current as _TreeTableEntryBase;
        } else if (cmp < 0) {
          current = branch.left;
        } else {
          current = branch.right;
        }
      }

      var leaf = current as _TreeTableEntryBase;

      if (comparer != null) {
        cmp = Comparable.compare(key, leaf.getSortKey());
      } else if (value.getMinimum() is Comparable) {
        cmp = comparableKey.compareTo(leaf.getSortKey());
      }

      comparableKey = null;

      if (cmp == 0) {
        return leaf;
      }

      _TreeTableBranchBase newBranch = leaf.createBranch(this);

      if (cmp < 0) {
        newBranch.setLeft(
            value, false, this.sorted); // will set leaf.Parent ...
        newBranch.right = leaf;
      } else if (cmp > 0) {
        newBranch.setLeft(leaf, false, this.sorted); // will set leaf.Parent ...
        newBranch.right = value;
      }

      if (branch == null) {
        this._root = newBranch;
      } else {
        this._replaceNode(branch, leaf, newBranch, false);
      }
      this.insertFixup(newBranch, inAddMode);
      return value;
    }
  }

  /// Finds a node in a sorted tree.
  ///
  /// * key - _required_ - Key needs to be find an index
  ///
  /// Returns the index of the key
  int indexOfKey(Object key) {
    _TreeTableEntryBase entry = this.findKey(key);
    if (entry == null) {
      return -1;
    }

    return entry.getPosition();
  }

  /// Finds a node in a sorted tree that matches the specified key.
  ///
  /// * key - _required_ - The key to search.
  ///
  /// Returns the node; `NULL` if not found.
  _TreeTableEntryBase findKey(Object key) {
    return this._findKey(key, false);
  }

  /// Finds the node in a sorted tree is just one entry ahead of the
  /// node with the specified key.
  ///
  /// It searches for the largest possible
  /// key that is smaller than the specified key.
  ///
  /// * key - _required_ - The key to search.
  ///
  /// Returns the node; `NULL` if not found.
  _TreeTableEntryBase findHighestSmallerOrEqualKey(Object key) {
    return this._findKey(key, true);
  }

  _TreeTableEntryBase _lastFoundEntry;
  Object _lastFoundEntryKey;
  bool _lastFoundEntryHighestSmallerValue = false;

  _TreeTableEntryBase cacheLastFoundEntry(
      _TreeTableEntryBase entry, Object key, bool highestSmallerValue) {
    this.lastIndex = -1;
    this._lastFoundEntry = entry;
    this._lastFoundEntryKey = key;
    this._lastFoundEntryHighestSmallerValue = highestSmallerValue;
    return this._lastFoundEntry;
  }

  _TreeTableEntryBase _findKey(Object key, bool highestSmallerValue) {
    if (!this.sorted) {
      throw Exception("This tree is not sorted.");
    }

    var comparableKey = key as Comparable;
    if (this.root == null) {
      // replace root
      return null;
    } else {
      Comparable comparer = this.comparer;
      int cmp = 0;

      if (this._lastFoundEntry != null &&
          this._lastFoundEntryKey != null &&
          key != null &&
          this._lastFoundEntryHighestSmallerValue == highestSmallerValue) {
        if (comparer != null) {
          cmp = Comparable.compare(key, this._lastFoundEntry.getMinimum());
        } else if (comparableKey != null) {
          cmp = comparableKey.compareTo(this._lastFoundEntry.getMinimum());
        }

        if (cmp == 0) {
          return this._lastFoundEntry;
        }
      }

      // find node
      _TreeTableBranchBase branch;
      _TreeTableNodeBase current = this.root;

      _TreeTableNodeBase lastLeft;

      while (!current.isEntry()) {
        branch = current as _TreeTableBranchBase;
        if (comparer != null) {
          cmp = Comparable.compare(key, branch.right.getMinimum());
        } else if (comparableKey != null) {
          cmp = comparableKey.compareTo(branch.right.getMinimum());
        } else {
          throw Exception("No Comparer specified.");
        }

        if (cmp == 0) {
          current = branch.right;
          while (!current.isEntry()) {
            current = (current as _TreeTableBranchBase).left;
          }

          return this.cacheLastFoundEntry(current, key, highestSmallerValue);
        } else if (cmp < 0) {
          current = branch.left;
          lastLeft = branch.left;
        } else {
          current = branch.right;
        }
      }

      var leaf = current as _TreeTableEntryBase;

      if (comparer != null) {
        cmp = Comparable.compare(key, leaf.getSortKey());
      } else if (comparableKey != null) {
        cmp = comparableKey.compareTo(leaf.getSortKey());
      }

      comparableKey = null;
      if (cmp == 0) {
        return this.cacheLastFoundEntry(leaf, key, highestSmallerValue);
      }

      if (highestSmallerValue) {
        if (cmp < 0) {
          return this.cacheLastFoundEntry(leaf, key, highestSmallerValue);
        } else if (lastLeft != null) {
          current = lastLeft;
          while (!current.isEntry()) {
            current = (current as _TreeTableBranchBase).right;
          }

          return this.cacheLastFoundEntry(current, key, highestSmallerValue);
        }
      }

      this._lastFoundEntry = null;
      return null;
    }
  }

  /// Gets a value indicating whether the tree is Read-only or not.
  bool get isReadOnly => false;

  /// Removes a node at the specified position.
  ///
  /// * index - _required_ - Index value
  void removeAt(int index) {
    this.remove(this[index]);
  }

  /// Inserts a node at the specified index.
  ///
  /// * index - _required_ - Position where to insert the value
  /// * value - _required_ - Tree value need to be insert
  @override
  void insert(int index, Object value) {
    this.insertBase(index, value as _TreeTableNodeBase);
  }

  /// Removes the node with the specified value.
  ///
  /// * value - _required_ - Value needs to be remove
  @override
  bool remove(Object value) {
    return this.removeBase(value as _TreeTableNodeBase);
  }

  /// Indicates whether the node belongs to this tree.
  ///
  /// * value - _required_ - Node value
  ///
  /// Returns the boolean value indicating whether the node belongs to this tree.
  @override
  bool contains(Object value) {
    return this.containsBase(value as _TreeTableNodeBase);
  }

  /// Clears all nodes in the tree.
  @override
  void clear() {
    this._root = null;
    this._lastAddBranch = null;
    this.lastIndex = -1;
    this._lastIndexLeaf = null;
    this.cacheLastFoundEntry(null, null, false);
  }

  /// Sets the index of the specified node.
  ///
  /// * value - _required_ - Tree value
  ///
  /// Returns the index of the specified node.
  @override
  int indexOf(Object value) {
    return this.indexOfBase(value as _TreeTableNodeBase);
  }

  /// Adds the specified node to the tree.
  ///
  /// * value - _required_ - Adding value
  ///
  /// Returns the zero-based collection index at which the value has been added
  @override
  int add(Object value) {
    return this.addBase(value as _TreeTableNodeBase);
  }

  /// Gets a value indicating whether the nodes can be added or removed.
  bool get isFixedSize => false;

  /// Gets a value indicating whether the tree is Synchronized or not.
  bool get isSynchronized => false;

  /// Gets the number of leaves.
  ///
  /// Returns the number of leaves.
  int getCount() {
    // I made this a function so that in debug Watch window the Count is not shown (and
    // tree gets unintentionally initialized)
    return this._root == null ? 0 : this._root.getCount();
  }

  /// Gets the number of leaves.
  int get count => this.countBase;

  /// Gets the number of leaves.
  int get countBase => this.getCount();

  /// Copies the element from this collection into an array.
  ///
  /// * array - _required_ - The destination array.
  /// * index - _required_ - The starting index in the destination array.
  @override
  void copyTo(List<dynamic> array, int index) {
    this.copyToBase(array, index);
  }

  /// Copies the elements from this collection into an array.
  ///
  /// * array - _required_ - The destination array.
  /// * index - _required_ - The starting index in the destination array.
  void copyToBase(List<_TreeTableNodeBase> array, int index) {
    int count = this.getCount();
    for (int i = 0; i < count; i++) {
      array[i + index] = this[i];
    }
  }

  /// Gets an object that can be used to synchronize access to the `ICollection`.
  ///
  /// Returns - _required_ - an object that can be used to synchronize access to the `ICollection`.
  Object get syncRoot => null;

  /// Gets an enumerator.
  ///
  /// Returns an enumerator.
  @override
  _EnumeratorBase getEnumerator() {
    return this.getEnumerator();
  }

  /// Gets a [TreeTableEnumerator].
  ///
  /// Returns a [TreeTableEnumerator].
  _TreeTableEnumerator getEnumeratorBase() {
    return new _TreeTableEnumerator(this);
  }
}

class _TreeTableEnumerator implements _EnumeratorBase //, IDisposable
{
  _TreeTableNodeBase _cursor, _next;
  _TreeTableBase _tree;

  /// Initializes a new instance of the `TreeTableEnumerator` class.
  ///
  /// * tree - _required_ - Tree instance
  _TreeTableEnumerator(_TreeTableBase tree) {
    this._tree = tree;
    this._cursor = null;
    if (tree.count > 0) {
      this._next = tree[0] as _TreeTableNodeBase;
    }
  }

  /// Resets the enumerator.
  @override
  void reset() {
    this._cursor = null;
    if (this._tree.count > 0) {
      this._next = this._tree[0] as _TreeTableNodeBase;
    } else {
      this._next = null;
    }
  }

  /// Gets the current enumerator.
  Object get current => this.current;

  /// Gets the current node.
  _TreeTableEntryBase get currentBase => this._cursor as _TreeTableEntryBase;

  /// Indicates whether to move to the next node.
  ///
  /// Returns a boolean value indicating whether to move to the next node.
  bool moveNext() {
    if (this._next == null) {
      return false;
    }

    this._cursor = this._next;

    _TreeTableBranchBase _parent = this._cursor.parent;

    if (_parent == null) {
      this._next = null;
      return true;
    } else {
      if (_MathHelper.referenceEquals(this._cursor, _parent.left)) {
        this._next = _parent.right;
      } else {
        _TreeTableBranchBase parentParent = _parent.parent;
        if (parentParent == null) {
          this._next = null;
          return true;
        } else {
          while (_MathHelper.referenceEquals(parentParent.right, _parent)) {
            _parent = parentParent;
            parentParent = parentParent.parent;
            if (parentParent == null) {
              this._next = null;
              return true;
            }
          }

          this._next = parentParent.right;
        }
      }

      while (!this._next.isEntry()) {
        this._next = (this._next as _TreeTableBranchBase).left;
      }
    }

    return this._cursor != null;
  }
}

/// An object that holds an [_TreeTableEntryBase].
class _TreeTableEntryBaseSource {
  /// Gets a reference to the [_TreeTableEntryBase].
  _TreeTableEntryBase get entry => _entry;
  _TreeTableEntryBase _entry;

  /// Sets a reference to the [_TreeTableEntryBase].
  set entry(_TreeTableEntryBase value) {
    if (value == _entry) return;
    _entry = value;
  }
}

/// A collection of `_TreeTableEntryBaseSource` objects
/// that are internally using a `ITreeTable`.
class _TreeTableEntrySourceCollection extends _ListBase {
  _TreeTableBase inner;

  /// Initializes a new instance of the `TreeTableEntrySourceCollection` class.
  _TreeTableEntrySourceCollection() {
    this.inner = _TreeTable(false);
  }

  ///Initializes a new instance of the `TreeTableEntrySourceCollection` class.
  ///
  /// * sorted - _required_ - Boolean value
  _TreeTableEntrySourceCollection.fromSorted(bool sorted) {
    this.inner = _TreeTable(sorted);
  }

  ///Initializes a new instance of the `TreeTableEntrySourceCollection` class.
  ///
  /// * inner - _required_ - Tree table
  _TreeTableEntrySourceCollection.fromInner(_TreeTableBase inner) {
    this.inner = inner;
  }

  /// Gets a value indicating whether the [BeginInit] was called or not.
  bool get isInitializing => this.inner.isInitializing;

  /// Optimizes insertion of many elements when tree is initialized for the first time.
  void beginInit() {
    this.inner.beginInit();
  }

  /// Ends optimization of insertion of elements when tree is initialized for the first time.
  void endInit() {
    this.inner.endInit();
  }

  /// Gets an `_TreeTableEntryBaseSource` at a specific position.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns the entry value for the specified position.
  operator [](index) {
    var entry = this.inner[index] as _TreeTableEntryBase;
    if (entry == null) {
      return null;
    }

    return entry.value as _TreeTableEntryBaseSource;
  }

  /// Sets an `_TreeTableEntryBaseSource` at a specific position.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns the entry value for the specified position.
  operator []=(index, value) {
    var entry = _TreeTableEntry()..value = value;
    value.entry = entry;
    this.inner[index] = entry;
  }

  /// Indicates whether object belongs to this collection.
  ///
  /// * value - _required_ - The value of the object.
  ///
  /// Returns `True` if object belongs to the collection. `false` otherwise.
  bool containsBase(_TreeTableEntryBaseSource value) {
    if (value == null) {
      return false;
    }

    return this.inner.contains(value.entry);
  }

  /// Returns the position of a object in the collection.
  /// * value - _required_ - The value of the object.
  /// Returns - _required_ - the position of the object.
  int indexOfBase(_TreeTableEntryBaseSource value) {
    return this.inner.indexOf(value.entry);
  }

  /// Copies the contents of the collection to an array.
  ///
  /// * array - _required_ - Destination array.
  /// * index - _required_ - Starting index of the destination array.
  void copyToBase(List<_TreeTableEntryBaseSource> array, int index) {
    int count = this.inner.count;
    for (int n = 0; n < count; n++) {
      array[index + n] = [n] as _TreeTableEntryBaseSource;
    }
  }

  /// Returns a strongly typed enumerator.
  _TreeTableEntrySourceCollectionEnumerator getEnumerator() {
    return _TreeTableEntrySourceCollectionEnumerator(this);
  }

  /// Inserts an object at the specified index.
  ///
  /// * index - _required_ - Index value where the object is to be inserted.
  /// * value - _required_ - Value of the object to insert.
  void insertBase(int index, _TreeTableEntryBaseSource value) {
    var entry = _TreeTableEntry();
    entry.value = value;
    value.entry = entry;
    this.inner.insert(index, entry);
  }

  /// Appends an object.
  ///
  /// * value - _required_ - The value of the object to append.
  ///
  /// Returns an instance for the tree with newly added entry.
  int addBase(_TreeTableEntryBaseSource value) {
    var entry = _TreeTableEntry();
    entry.value = value;
    value.entry = entry;
    return this.inner.add(entry);
  }

  /// Removes the object.
  ///
  /// * value - _required_ - The value of the object to remove.
  void removeBase(_TreeTableEntryBaseSource value) {
    this.inner.remove(value.entry);
  }

  /// Gets a value indicating whether tree is Read-only or not.
  bool get isReadOnly => false;

  /// Removes a node at the specified index.
  ///
  /// * index - _required_ - Index value of the node to remove.
  void removeAtBase(int index) {
    this.inner.removeAt(index);
  }

  /// Inserts the object at the specified index.
  ///
  /// * index - _required_ - Index value of the object to insert.
  /// * value - _required_ - Value of the object to insert.
  @override
  void insert(int index, Object value) {
    this.insertBase(index, value as _TreeTableEntryBaseSource);
  }

  /// Removes the specified object.
  ///
  /// * value - _required_ - Value of the object to remove.
  @override
  void remove(Object value) {
    this.removeBase(value as _TreeTableEntryBaseSource);
  }

  /// Indicate whether the specified object belongs to this collection.
  ///
  /// * value - _required_ - Object value to look for.
  /// Returns - _required_ - true if object belongs to the collection; false otherwise.
  @override
  bool contains(Object value) {
    return this.containsBase(value as _TreeTableEntryBaseSource);
  }

  /// Clears all nodes in the tree.
  void clear() {
    this.inner.clear();
  }

  /// Returns the index of the specified object.
  ///
  /// * value - _required_ - Value of the object.
  ///
  /// Returns Index value of the object.
  @override
  int indexOf(Object value) {
    return this.indexOfBase(value as _TreeTableEntryBaseSource);
  }

  /// Adds the specified object to the collection.
  ///
  /// * value - _required_ - Value of the object to add.
  ///
  /// Returns the zero-based collection index at which the value has been added
  @override
  int add(Object value) {
    return this.addBase(value as _TreeTableEntryBaseSource);
  }

  /// Gets a value indicating whether the nodes can be added or removed.
  bool get isFixedSize => false;

  /// Gets a value indicating whether the tree is Synchronized or not.
  bool get isSynchronized => false;

  /// Gets the number of objects in this collection.
  int get count => this.inner.count;

  /// Copies elements to destination array.
  ///
  /// * array - _required_ - Destination array.
  /// * index - _required_ - Starting index of the destination array.
  @override
  void copyTo(List<dynamic> array, int index) {
    this.copyToBase(array, index);
  }
}

/// A strongly typed enumerator for the `TreeTableEntrySourceCollection`.
class _TreeTableEntrySourceCollectionEnumerator implements _EnumeratorBase {
  _TreeTableEnumerator inner;

  /// Initializes a new instance of the `TreeTableEntrySourceCollectionEnumerator` class.
  ///
  /// * collection - _required_ - Collection value
  _TreeTableEntrySourceCollectionEnumerator(
      _TreeTableEntrySourceCollection collection) {
    this.inner = _TreeTableEnumerator(collection.inner);
  }

  /// Resets the enumerator.
  @override
  void reset() {
    this.inner.reset();
  }

  /// Gets the current enumerator.
  Object get current => this.currentBase;

  /// Gets the current `_TreeTableEntryBaseSource` object.
  _TreeTableEntryBaseSource get currentBase =>
      this.inner.currentBase.value as _TreeTableEntryBaseSource;

  /// Indicates whether to move to the next object in the collection.
  ///
  /// Returns the boolean value indicates whether to move to the next object in the collection.
  bool moveNext() {
    return this.inner.moveNext();
  }
}
