part of datagrid;

/// Interface definition for a node that has counters and summaries.
mixin _TreeTableCounterNodeBase on _TreeTableSummaryNodeBase {
  /// The total of this node's counter and child nodes.
  ///
  /// Returns the total of this node's counter and child nodes (cached).
  _TreeTableCounterBase getCounterTotal();

  /// Gets the cumulative position of this node.
  ///
  /// Returns  Returns the cumulative position of this node.
  _TreeTableCounterBase get getCounterPosition;

  /// Marks all counters dirty in this node and child nodes.
  ///
  /// * notifyCounterSource - _required_ - If set to `true` notify counter source.
  void invalidateCounterTopDown(bool notifyCounterSource);
}

/// Interface definition for an object that has counters.
abstract class _TreeTableCounterSourceBase {
  /// Gets the counter object with counters.
  ///
  /// Returns the counter object with counters.
  _TreeTableCounterBase getCounter();

  /// Marks all counters dirty in this object and child nodes.
  ///
  /// * notifyCounterSource - _required_ - If set to true notify counter source.
  void invalidateCounterTopDown(bool notifyCounterSource);

  /// Marks all counters dirty in this object and parent nodes.
  void invalidateCounterBottomUp();
}

/// Interface definition for a counter object.
abstract class _TreeTableCounterBase {
  /// Combines this counter object with another counter and returns a new object. A cookie can specify
  /// a specific counter type.
  ///
  /// * other - _required_ - Counter total
  /// * cookie - _required_ - Cookie value.
  ///
  /// Returns the new object
  _TreeTableCounterBase combine(_TreeTableCounterBase other, int cookie);

  /// Compares this counter with another counter. A cookie can specify
  /// a specific counter type.
  ///
  /// * other - _required_ - The other.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the compared value.
  double compare(_TreeTableCounterBase other, int cookie);

  /// Indicates whether the counter object is empty. A cookie can specify
  /// a specific counter type.
  ///
  /// * cookie - _required_ - The cookie.
  /// Returns `true` if the specified cookie is empty; otherwise,`false`.
  bool isEmpty(int cookie);

  /// Gets the integer value of the counter. A cookie specifies
  /// a specific counter type.
  ///
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the integer value of the counter.
  double getValue(int cookie);

  /// Gets the Counter Kind.
  ///
  /// Returns the kind.
  int get kind => _kind;
  int _kind;
}

/// Default counter cookies for identifying counter types.
class _TreeTableCounterCookies {
  /// All counters.
  static int countAll = 0xffff;

  /// Visible Counter.
  static int countVisible = 0x8000;
}

/// A tree table branch with a counter.
class _TreeTableWithCounterBranch extends _TreeTableWithSummaryBranch
    with _TreeTableCounterNodeBase {
  _TreeTableCounterBase counter;

  ///Initializes a new instance of the `_TreeTableWithCounterBranch` class.
  ///
  /// * tree - _required_ - Tree instance
  _TreeTableWithCounterBranch(_TreeTable tree) : super(tree);

  /// Gets the tree this branch belongs to.
  _TreeTableWithCounter get treeTableWithCounter {
    return this.tree as _TreeTableWithCounter;
  }

  /// Gets the parent branch.
  @override
  _TreeTableWithCounterBranch get parent =>
      super.parent as _TreeTableWithCounterBranch;
  set parent(var value) {
    super.parent = value;
  }

  /// Gets the cumulative counter position object of a child node with all counter values.
  ///
  /// * node - _required_ - The node.
  ///
  /// Returns the cumulative counter position object of a child node with all counter values.
  _TreeTableCounterBase getCounterPositionOfChild(_TreeTableNodeBase node) {
    _TreeTableCounterBase pos = this.getCounterPosition;

    if (_MathHelper.referenceEquals(node, this.right)) {
      return pos.combine(this.getLeftNode().getCounterTotal(),
          _TreeTableCounterCookies.countAll);
    } else if (_MathHelper.referenceEquals(node, this.left)) {
      return pos;
    }

    throw ArgumentError("must be a child node");
  }

  /// The left branch node cast to ITreeTableCounterNode.
  ///
  /// Returns the left branch node cast to ITreeTableCounterNode.
  _TreeTableCounterNodeBase getLeftNode() {
    return left as _TreeTableCounterNodeBase;
  }

  /// The right branch node cast to ITreeTableCounterNode.
  ///
  /// Returns the right branch node cast to ITreeTableCounterNode.
  _TreeTableCounterNodeBase getRightNode() {
    return right as _TreeTableCounterNodeBase;
  }

  /// Gets the total of this node's counter and child nodes (cached).
  ///
  /// Returns  Returns the total of this node's counter and child nodes (cached).
  _TreeTableCounterBase getCounterTotal() {
    if (tree.isInitializing) {
      return null;
    } else if (this.counter == null) {
      _TreeTableCounterBase _left = this.getLeftNode().getCounterTotal();
      _TreeTableCounterBase _right = this.getRightNode().getCounterTotal();
      if (_left != null && _right != null) {
        this.counter = _left.combine(_right, _TreeTableCounterCookies.countAll);
      }
    }

    return this.counter;
  }

  /// Gets the cumulative position of this node.
  ///
  /// Returns the cumulative position of this node.
  _TreeTableCounterBase get getCounterPosition {
    if (this.parent == null) {
      return this.treeTableWithCounter.getStartCounterPosition();
    }

    return this.parent.getCounterPositionOfChild(this);
  }

  /// Invalidates the counter bottom up.
  ///
  /// * notifyCounterSource - _required_ - If set to true notify counter source.
  @override
  void invalidateCounterBottomUp(bool notifyCounterSource) {
    if (tree.isInitializing) {
      return;
    }

    this.counter = null;
    if (this.parent != null) {
      this.parent.invalidateCounterBottomUp(notifyCounterSource);
    } else if (notifyCounterSource) {
      if (tree is _TreeTableWithCounter) {
        _TreeTableCounterSourceBase tcs;
        tcs = tree.tag as _TreeTableCounterSourceBase;
        if (tcs != null) {
          tcs.invalidateCounterBottomUp();
        }

        tcs = (tree as _TreeTableWithCounter).parentCounterSource;
        if (tcs != null) {
          tcs.invalidateCounterBottomUp();
        }
      }
    }
  }

  /// Marks all counters dirty in this node and child nodes.
  ///
  /// * notifyCounterSource - _required_ - If set to true notify counter source.
  void invalidateCounterTopDown(bool notifyCounterSource) {
    if (tree.isInitializing) {
      return;
    }

    this.counter = null;
    this.getLeftNode().invalidateCounterTopDown(notifyCounterSource);
    this.getRightNode().invalidateCounterTopDown(notifyCounterSource);
  }

  /// The left branch node cast to ITreeTableCounterNode.
  ///
  /// Returns the left branch node cast to ITreeTableCounterNode.
  _TreeTableCounterNodeBase getLeftC() {
    return this.getLeftNode();
  }

  /// The right branch node cast to ITreeTableCounterNode.
  ///
  /// Returns the right branch node cast to ITreeTableCounterNode.
  _TreeTableCounterNodeBase getRightC() {
    return this.getRightNode();
  }
}

/// A tree leaf with value, sort key and counter information.
class _TreeTableWithCounterEntry extends _TreeTableWithSummaryEntryBase
    with _TreeTableCounterNodeBase {
  _TreeTableCounterBase _counter;

  /// Gets the tree this leaf belongs to.
  _TreeTableWithCounter get treeTableWithCounter {
    return super.tree as _TreeTableWithCounter;
  }

  /// Gets the parent branch.
  _TreeTableWithCounterBranch get parent {
    return super.parent as _TreeTableWithCounterBranch;
  }

  /// Sets the parent branch.
  set parent(var value) {
    super.parent = value;
  }

  /// Creates a branch that can hold this entry when new leaves are inserted into the tree.
  ///
  /// * tree - _required_ - Tree instance
  ///
  /// Returns the instance of newly created branch
  @override
  _TreeTableBranchBase createBranch(_TreeTable tree) {
    return _TreeTableWithCounterBranch(tree);
  }

  /// Gets the cumulative position of this node.
  ///
  /// Returns  Returns the cumulative position of this node.
  @override
  _TreeTableCounterBase get getCounterPosition {
    if (this.parent == null) {
      if (treeTableWithCounter == null) {
        return null;
      }

      return treeTableWithCounter.getStartCounterPosition();
    }

    return this.parent.getCounterPositionOfChild(this);
  }

  /// Gets the value as `_TreeTableCounterSourceBase`.
  ///
  /// Returns the value as `_TreeTableCounterSourceBase`.
  _TreeTableCounterSourceBase getCounterSource() {
    return value as _TreeTableCounterSourceBase;
  }

  /// Indicates whether the counter was set dirty.
  ///
  /// Returns `True` if dirty; `False` otherwise.
  bool isCounterDirty() {
    return this._counter == null;
  }

  /// Gets the total of this node's counter and child nodes.
  ///
  /// Returns the total of this node's counter and child nodes.
  _TreeTableCounterBase getCounterTotal() {
    if (this._counter == null) {
      _TreeTableCounterSourceBase source = this.getCounterSource();
      if (source != null) {
        this._counter = source.getCounter();
      }
    }

    return this._counter;
  }

  /// Reset cached counter.
  void invalidateCounter() {
    this._counter = null;
  }

  /// Invalidates the counter bottom up.
  ///
  /// * notifyCounterSource - _required_ - If set to true notify counter source.
  @override
  void invalidateCounterBottomUp(bool notifyCounterSource) {
    this._counter = null;
    if (this.parent != null) {
      this.parent.invalidateCounterBottomUp(notifyCounterSource);
    } else if (notifyCounterSource) {
      if (tree is _TreeTableWithCounter) {
        var tcs = tree.tag as _TreeTableCounterSourceBase;
        if (tcs != null) {
          tcs.invalidateCounterBottomUp();
        }

        tcs = (tree as _TreeTableWithCounter).parentCounterSource;
        if (tcs != null) {
          tcs.invalidateCounterBottomUp();
        }
      }
    }
  }

  /// Marks all summaries dirty in this node and child nodes.
  ///
  /// * notifyCounterSource - _required_ - If set to true notify counter source.
  void invalidateCounterTopDown(bool notifyCounterSource) {
    this._counter = null;
    if (notifyCounterSource) {
      _TreeTableCounterSourceBase source = this.getCounterSource();
      if (notifyCounterSource && source != null) {
        source.invalidateCounterTopDown(notifyCounterSource);
      }
    }
  }
}

/// A balanced tree with [_TreeTableWithCounterEntry] entries.
class _TreeTableWithCounter extends _TreeTableWithSummary {
  _TreeTableCounterBase _startPos;

  /// Initializes a new instance of the [TreeTableWithCounter] class.
  ///
  /// * startPosition - _required_ - Sorting position
  /// * sorted - _required_ - Boolean value
  _TreeTableWithCounter(_TreeTableCounterBase startPosition, bool sorted)
      : super(sorted) {
    this._startPos = startPosition;
  }

  /// Gets an object that implements the [Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.Distances] property.
  _TreeTableCounterSourceBase get parentCounterSource => _parentCounterSource;
  _TreeTableCounterSourceBase _parentCounterSource;

  /// Sets an object that implements the [Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.Distances] property.
  set parentCounterSource(_TreeTableCounterSourceBase value) {
    if (value == _parentCounterSource) return;
    _parentCounterSource = value;
  }

  /// Ends optimization of insertion of elements when tree is initialized for the first time.
  @override
  void endInit() {
    super.endInit();
  }

  /// Marks all counters dirty.
  ///
  /// * notifyCounterSource - _required_ - Boolean value
  void invalidateCounterTopDown(bool notifyCounterSource) {
    if (this.root != null) {
      (this.root as _TreeTableCounterNodeBase)
          .invalidateCounterTopDown(notifyCounterSource);
    }
  }

  /// Gets the total of all counters in this tree.
  ///
  /// Returns the total of all counters in this tree.
  _TreeTableCounterBase getCounterTotal() {
    if (this.root == null) {
      return this._startPos;
    }

    return (this.root as _TreeTableCounterNodeBase).getCounterTotal();
  }

  /// Gets the starting counter for this tree.
  ///
  /// Returns  Returns the starting counter for this tree.
  _TreeTableCounterBase getStartCounterPosition() {
    return this._startPos;
  }

  /// Overloaded. Gets an entry at the specified counter position. A cookie defines the type of counter.
  ///
  /// * searchPosition - _required_ - The search position.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns an entry at the specified counter position.
  _TreeTableWithCounterEntry getEntryAtCounterPosition(
      _TreeTableCounterBase searchPosition, int cookie) {
    return this.getEntryAtCounterPositionWithForParameter(
        this.getStartCounterPosition(), searchPosition, cookie, false);
  }

  /// Gets an entry at the specified counter position. A cookie defines the type of counter.
  ///
  /// * searchPosition - _required_ - The search position.
  /// * cookie - _required_ - The cookie.
  /// * preferLeftMost - _required_ - Indicates if the leftmost entry should be returned if multiple tree elements have the
  /// same SearchPosition.
  ///
  /// Returns an entry at the specified counter position.
  _TreeTableWithCounterEntry getEntryAtCounterPositionwithThreeParameter(
      _TreeTableCounterBase searchPosition, int cookie, bool preferLeftMost) {
    return this.getEntryAtCounterPositionWithForParameter(
        this.getStartCounterPosition(), searchPosition, cookie, preferLeftMost);
  }

  /// Gets the entry at counter position.
  ///
  /// * start - _required_ - The start.
  /// * searchPosition - _required_ - The search position.
  /// * cookie - _required_ - The cookie.
  /// * preferLeftMost - _required_ - If set to true prefer left most.
  ///
  /// Returns an entry at the specified counter position.
  _TreeTableWithCounterEntry getEntryAtCounterPositionWithForParameter(
      _TreeTableCounterBase start,
      _TreeTableCounterBase searchPosition,
      int cookie,
      bool preferLeftMost) {
    if (searchPosition.compare(this.getStartCounterPosition(), cookie) < 0) {
      throw Exception("SearchPosition");
    }

    if (searchPosition.compare(this.getCounterTotal(), cookie) > 0) {
      throw Exception("$searchPosition out of range ${this.getCounterTotal()}");
    }

    if (this.root == null) {
      return null;
    } else {
      // find node
      _TreeTableNodeBase currentNode = this.root;
      _TreeTableCounterBase currentNodePosition = start;
      return this.getEntryAtCounterPositionWithSixParameter(currentNode, start,
          searchPosition, cookie, preferLeftMost, currentNodePosition);
    }
  }

  /// An object that implements the `Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.Distances` property.
  ///
  /// * currentNode - _required_ - Current node value
  /// * start - _required_ - Start counter
  /// * searchPosition - _required_ - Position needs to be search
  /// * cookie - _required_ - Cookie value
  /// * preferLeftMost - _required_ - Indicates if the leftmost entry should be returned if multiple tree elements have the
  /// same SearchPosition
  /// * currentNodePosition - _required_ - Position of the current node
  ///
  /// Returns the current node.
  _TreeTableWithCounterEntry getEntryAtCounterPositionWithSixParameter(
      _TreeTableNodeBase currentNode,
      _TreeTableCounterBase start,
      _TreeTableCounterBase searchPosition,
      int cookie,
      bool preferLeftMost,
      _TreeTableCounterBase currentNodePosition) {
    _TreeTableWithCounterBranch savedBranch;
    currentNodePosition = start;
    while (!currentNode.isEntry()) {
      var branch = currentNode as _TreeTableWithCounterBranch;
      var leftB = branch.left as _TreeTableCounterNodeBase;
      _TreeTableCounterBase rightNodePosition =
          currentNodePosition.combine(leftB.getCounterTotal(), cookie);

      if (searchPosition.compare(rightNodePosition, cookie) < 0) {
        currentNode = branch.left;
      } else if (preferLeftMost &&
          searchPosition.compare(currentNodePosition, cookie) == 0) {
        while (!currentNode.isEntry()) {
          branch = currentNode as _TreeTableWithCounterBranch;
          currentNode = branch.left;
        }
      } else {
        if (preferLeftMost &&
            searchPosition.compare(rightNodePosition, cookie) == 0) {
          _TreeTableCounterBase currentNode2Position;
          _TreeTableNodeBase currentNode2 = this
              .getEntryAtCounterPositionWithSixParameter(
                  branch.left,
                  currentNodePosition,
                  searchPosition,
                  cookie,
                  preferLeftMost,
                  currentNode2Position);
          if (rightNodePosition.compare(currentNode2Position, cookie) == 0) {
            currentNode = currentNode2;
            currentNodePosition = currentNode2Position;
          } else {
            currentNodePosition = rightNodePosition;
            currentNode = branch.right;
          }
        } else {
          if (savedBranch == null) {
            savedBranch = branch;
          }
          currentNodePosition = rightNodePosition;
          currentNode = branch.right;
        }
      }
    }

    return currentNode as _TreeTableWithCounterEntry;
  }

  /// Gets the subsequent entry in the collection for which the specific counter is not empty.
  /// A cookie defines the type of counter.
  ///
  /// * current - _required_ - The current.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the subsequent entry in the collection for which the specific counter is not empty.
  _TreeTableEntryBase getNextNotEmptyCounterEntry(
      _TreeTableEntryBase current, int cookie) {
    _TreeTableBranchBase parent = current.parent;
    _TreeTableNodeBase next;

    if (parent == null) {
      next = null;
      return null;
    } else {
      next = current;
      // walk up until we find a branch that has visible entries
      do {
        if (_MathHelper.referenceEquals(next, parent.left)) {
          next = parent.right;
        } else {
          _TreeTableBranchBase parentParent = parent.parent;
          if (parentParent == null) {
            return null;
          } else {
            while (_MathHelper.referenceEquals(parentParent.right, parent)
                // for something that most likely went wrong when
                // adding the node or when doing a rotation ...
                ||
                _MathHelper.referenceEquals(parentParent.right, next)) {
              parent = parentParent;
              parentParent = parentParent.parent;
              if (parentParent == null) {
                return null;
              }
            }

            if (next == parentParent.right) {
              throw new Exception();
            } else {
              next = parentParent.right;
            }
          }
        }
      } while (next != null &&
          (next as _TreeTableCounterNodeBase)
              .getCounterTotal()
              .isEmpty(cookie));

      // walk down to most left leaf that has visible entries
      while (!next.isEntry()) {
        var branch = next as _TreeTableBranchBase;
        next = !(branch.left as _TreeTableCounterNodeBase)
                .getCounterTotal()
                .isEmpty(cookie)
            ? branch.left
            : branch.right;
      }
    }

    return next as _TreeTableEntryBase;
  }

  /// Returns the previous entry in the collection for which the specific counter is not empty.
  ///
  /// * current - _required_ - The current.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the previous entry in the collection for which the specific counter is not empty.
  _TreeTableEntryBase getPreviousNotEmptyCounterEntry(
      _TreeTableEntryBase current, int cookie) {
    _TreeTableBranchBase parent = current.parent;
    _TreeTableNodeBase next;

    if (parent == null) {
      next = null;
      return null;
    } else {
      next = current;
      // walk up until we find a branch that has visible entries
      do {
        if (_MathHelper.referenceEquals(next, parent.right)) {
          next = parent.left;
        } else {
          _TreeTableBranchBase parentParent = parent.parent;
          if (parentParent == null) {
            return null;
          } else {
            while (_MathHelper.referenceEquals(parentParent.left, parent)
                // for something that most likely went wrong when
                // adding the node or when doing a rotation ...
                ||
                _MathHelper.referenceEquals(parentParent.left, next)) {
              parent = parentParent;
              parentParent = parentParent.parent;
              if (parentParent == null) {
                return null;
              }
            }
            if (next == parentParent.left) {
              throw Exception();
            } else {
              next = parentParent.left;
            }
          }
        }
      } while (next != null &&
          (next as _TreeTableCounterNodeBase)
              .getCounterTotal()
              .isEmpty(cookie));

      // walk down to most left leaf that has visible entries
      while (!next.isEntry()) {
        var branch = next as _TreeTableBranchBase;
        next = !(branch.right as _TreeTableCounterNodeBase)
                .getCounterTotal()
                .isEmpty(cookie)
            ? branch.right
            : branch.left;
      }
    }

    return next as _TreeTableEntryBase;
  }

  /// Gets the next entry in the collection for which CountVisible counter is not empty.
  ///
  /// * current - _required_ - The current.
  ///
  /// Returns the next entry in the collection for which CountVisible counter is not empty.
  _TreeTableWithCounterEntry getNextVisibleEntry(
      _TreeTableWithCounterEntry current) {
    return this.getNextNotEmptyCounterEntry(
        current, _TreeTableCounterCookies.countVisible);
  }

  /// Gets the previous entry in the collection for which CountVisible counter is not empty.
  ///
  /// * current - _required_ - The current.
  ///
  /// Returns the previous entry in the collection for which CountVisible counter is not empty.
  _TreeTableWithCounterEntry getPreviousVisibleEntry(
      _TreeTableWithCounterEntry current) {
    return this.getPreviousNotEmptyCounterEntry(
        current, _TreeTableCounterCookies.countVisible);
  }

  /// Gets a _TreeTableWithCounterEntry.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns a new instance for _TreeTableWithCounterEntry
  operator [](index) {
    return super[index] as _TreeTableWithCounterEntry;
  }

  /// Sets a _TreeTableWithCounterEntry.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns a new instance for _TreeTableWithCounterEntry
  operator []=(index, value) {
    super[index] = value;
  }

  /// Inserts a `_TreeTableWithCounterEntry` object at the specified index.
  ///
  /// * index - _required_ - The index.
  /// * value - _required_ - The value.
  @override
  void insert(int index, Object value) {
    super.insert(index, value);
  }

  /// Removes an object from the tree.
  ///
  /// * value - _required_ - The value.
  ///
  /// Returns the collection after removing the specified item from the tree collection
  @override
  bool remove(Object value) {
    return super.remove(value);
  }

  /// Indicates whether an entry belongs to the tree.
  ///
  /// * value - _required_ - The entry.
  ///
  /// Returns `true` if tree contains the specified entry; otherwise, `false`.
  @override
  bool contains(Object value) {
    if (value == null) {
      return false;
    }

    return super.contains(value);
  }

  /// Gets the position of an object in the tree.
  ///
  /// * value - _required_ - The value.
  ///
  /// Returns the position of an object in the tree.
  @override
  int indexOf(Object value) {
    return super.indexOf(value);
  }

  /// Appends an object.
  ///
  /// * value - _required_ - The value.
  ///
  /// Returns the zero-based collection index at which the value has been added.
  @override
  int add(Object value) {
    return super.add(value);
  }

  /// Copies the elements of this tree to an array.
  ///
  /// * array - _required_ - The array.
  /// * index - _required_ - The index.
  void copyTo(List<dynamic> array, int index) {
    super.copyTo(array, index);
  }
}
