part of datagrid;

/// A collection that manages lines with varying height and hidden state.
///
/// It has properties for header and footer lines, total line count, default
/// size of a line and also lets you add nested collections.
///
/// _Note:_ This is common for header, footer, total line count and default size of line.
class _LineSizeCollection extends _PaddedEditableLineSizeHostBase
    with _DistancesHostBase, _NestedDistancesHostBase {
  int _lineCount = 0;
  double _defaultLineSize = 1.0;
  _SortedRangeValueList<double> _lineSizes =
      _SortedRangeValueList<double>.from(-1);
  _SortedRangeValueList<bool> _lineHidden =
      _SortedRangeValueList<bool>.from(false);
  Map<int, _LineSizeCollection> _lineNested = Map<int, _LineSizeCollection>();
  int _headerLineCount = 0;
  int _footerLineCount = 0;
  _DistanceCounterCollectionBase _distances;

  ///Initializes a new instance of the [LineSizeCollection] class.
  _LineSizeCollection();

  /// Gets the total extent which is the total of all line sizes.
  ///
  /// Returns the total extent which is the total of all line sizes or [double.NaN].
  ///
  /// _Note:_ This property only
  /// works if the [DistanceCounterCollection] has been setup for pixel scrolling,
  /// otherwise it returns [double.NaN].
  double get totalExtent {
    // This only works if the DistanceCollection has been setup for pixel scrolling.
    if (this.distances.defaultDistance == this._defaultLineSize) {
      return this._distances.totalDistance;
    }
    return double.nan;
  }

  double _paddingDistance = 0;

  /// Gets the padding distance for the line.
  ///
  /// Returns the padding distance for the line.
  double get paddingDistance => this._paddingDistance;

  /// sets the padding distance for the line.
  set paddingDistance(double value) {
    if (this._paddingDistance != value) {
      this._paddingDistance = value;

      if (this.isSuspendUpdates) {
        return;
      }

      this._distances = _DistanceRangeCounterCollection.fromPaddingDistance(
          this.paddingDistance);
      this.initializeDistances();
    }
  }

  /// Gets the header line count.
  ///
  /// Returns the header line count.
  int get headerLineCount => this._headerLineCount;

  /// Sets the header line count.
  set headerLineCount(int value) {
    if (this._headerLineCount != value) {
      this._headerLineCount = value;
      if (this.onHeaderLineCountChanged != null) {
        this.onHeaderLineCountChanged();
      }
    }
  }

  /// Gets the footer line count.
  ///
  /// Returns the footer line count.
  int get footerLineCount => this._footerLineCount;

  /// Sets the footer line count.
  set footerLineCount(int value) {
    if (this._footerLineCount != value) {
      this._footerLineCount = value;
      if (this.onFooterLineCountChanged != null) {
        this.onFooterLineCountChanged();
      }
    }
  }

  /// Gets the line count.
  ///
  /// Returns the line count.
  int get lineCount => this._lineCount;

  /// Sets the line count.
  set lineCount(int value) {
    if (this._lineCount != value) {
      this._lineCount = value;

      if (this.isSuspendUpdates) {
        return;
      }

      if (this._distances != null) {
        this._distances.count = this._lineCount;
      }
    }
  }

  /// Gets the default size of lines.
  ///
  /// Returns the default size of lines.
  double get defaultLineSize => this._defaultLineSize;

  /// Sets the default size of lines.
  set defaultLineSize(double value) {
    if (this._defaultLineSize != value) {
      double savedValue = this._defaultLineSize;
      this._defaultLineSize = value;

      if (this.isSuspendUpdates) {
        return;
      }

      if (this._distances != null) {
        this.initializeDistances();
      }

      if (this.onDefaultLineSizeChanged != null) {
        var defaultLineSizeChangedArgs = _DefaultLineSizeChangedArgs.fromArgs(
            savedValue, this._defaultLineSize);
        this.onDefaultLineSizeChanged(defaultLineSizeChangedArgs);
      }
    }
  }

  /// Sets the line size for the range of lines.
  ///
  /// * from - _required_ - The start index of the line for which the line size is to be set.
  /// * to - _required_ - The end index of the line for which the line size is to be set.
  /// * size - _required_ - The line size to be set to the given range of lines.
  void setRange(int from, int to, double size) {
    int count = 0;
    double savevalue = this.getRange(from, count)[0];
    count = this.getRange(from, count)[1];
    this._lineSizes.setRange(from, to - from + 1, size);

    if (this.isSuspendUpdates) {
      return;
    }

    if (this._distances != null) {
      _DistancesUtil.distancesLineHiddenChanged(
          this._distances, this, from, to);
    }
    _RangeChangedArgs rangeChangedArgs;
    if (this.onLineSizeChanged != null) {
      rangeChangedArgs =
          _RangeChangedArgs.fromRangeChangedArgs(from, to, savevalue, size);
      this.onLineSizeChanged(rangeChangedArgs);
    }
  }

  /// Gets the line size at the specified index.
  ///
  /// * index - _required_ - Index value
  ///
  /// Returns the line size at the specified index.

  operator [](index) {
    int repeatValueCount;
    return this.getRange(index, repeatValueCount);
  }

  /// Sets the line size at the specified index.
  operator []=(index, value) {
    this.setRange(index, index, value);
  }

  List getRange(int index, int repeatValueCount) {
    repeatValueCount = 1;

    if (this._lineNested.containsKey(index)) {
      return [this._lineNested[index].totalExtent, repeatValueCount];
    }

    bool hide = this._lineHidden.getRange(index, repeatValueCount)[0];
    repeatValueCount = this._lineHidden.getRange(index, repeatValueCount)[1];
    if (hide) {
      return [0.0, repeatValueCount];
    }

    double size = this._lineSizes.getRange(index, repeatValueCount)[0];
    repeatValueCount = this._lineSizes.getRange(index, repeatValueCount)[1];
    if (size >= 0) {
      return [size, repeatValueCount];
    }
    return [this._defaultLineSize, repeatValueCount];
  }

  /// Sets the hidden state for the given range of lines.
  ///
  /// * from - _required_ - The start index of the line for which the hidden state to be set.
  /// * to - _required_ - The end index of the line for which the hidden state to be set.
  /// * hide - _required_ - A boolean value indicating whether to hide the lines. If set to true hide the lines.
  void setHidden(int from, int to, bool hide) {
    this._lineHidden.setRange(from, to - from + 1, hide);

    if (this.isSuspendUpdates) {
      return;
    }
    // DistancesLineHiddenChanged checks both hidden state and sizes together ...
    if (this._distances != null) {
      _DistancesUtil.distancesLineHiddenChanged(this.distances, this, from, to);
    }
    _HiddenRangeChangedArgs hiddenRangeChangedArgs;
    if (this.onLineHiddenChanged != null) {
      hiddenRangeChangedArgs = _HiddenRangeChangedArgs.fromArgs(from, to, hide);
      this.onLineHiddenChanged(hiddenRangeChangedArgs);
    }
  }

  /// Gets a value indicating whether the host supports nesting.
  ///
  /// Returns a boolean value indicating whether the host supports nesting.
  bool get supportsNestedLines => true;

  /// Gets the nested lines at the given index.
  ///
  /// * index - _required_ - The index at which the nested lines is to be obtained.
  ///
  /// Returns the [IEditableLineSizeHost] representing the nested lines.
  _EditableLineSizeHostBase getNestedLines(int index) {
    if (this._lineNested.containsKey(index)) {
      return this._lineNested[index];
    }

    return null;
  }

  /// Sets the nested lines at the given index.
  ///
  /// * index - _required_ - The index at which the nested lines is to be added.
  /// * nestedLines - _required_ - The nested lines to be added. If parameter is null the line will
  /// be converted to a normal (not nested) line with default line size.
  void setNestedLines(int index, _EditableLineSizeHostBase nestedLines) {
    if (nestedLines != null) {
      this._lineSizes[index] =
          -2; // -1 indicates default value, -2 indicates nested.
      this._lineNested[index] = nestedLines as _LineSizeCollection;
    } else {
      this._lineSizes[index] =
          -1; // -1 indicates default value, -2 indicates nested.
      this._lineNested.remove(index);
    }
    if (this.isSuspendUpdates) {
      return;
    }

    if (this._distances != null) {
      _DistancesUtil.distancesLineSizeChanged(
          this.distances, this, index, index);
    }

    if (this.onLineSizeChanged != null) {
      this.onLineSizeChanged(_RangeChangedArgs.fromArgs(index, index));
    }
  }

  /// Reset the lines with default line size.
  void resetLines() {
    this._lineSizes.clear();
  }

  /// Reset the line to become a normal (not nested) line with default line size.
  void resetNestedLines() {
    for (int i = 0; i < _lineNested.length; i++)
      this._lineNested.forEach((key, value) {
        this._lineSizes[key] = -1;
      });

    this._lineNested.clear();
  }

  /// Reset the line to become a normal (not nested) line with default line size.
  ///
  /// * index - _required_ - The index of the line to be reset.
  void resetNestedLinesWithArgs(int index) {
    this.setNestedLines(index, null);
  }

  /// Gets the default line size.
  ///
  /// Returns the default line size.
  double getDefaultLineSize() {
    return this.defaultLineSize;
  }

  /// Gets the line count.
  ///
  /// Returns the line count.
  int getLineCount() {
    return lineCount;
  }

  /// Gets the size of the line at the given index.
  ///
  /// * index - _required_ - The index of the line for which the size is to be obtained.
  /// * repeatValueCount - _required_ - The number of subsequent values with same size.
  ///
  /// Returns the size of the line at the given index.
  @override
  List getSize(int index, int repeatValueCount) {
    repeatValueCount = 1;
    _EditableLineSizeHostBase nested = this.getNestedLines(index);
    if (nested != null) {
      return [nested.totalExtent, repeatValueCount];
    }

    return this.getRange(index, repeatValueCount);
  }

  /// Gets the header line count.
  ///
  /// Returns the header line count.
  int getHeaderLineCount() {
    return this.headerLineCount;
  }

  /// Gets the footer line count.
  ///
  /// Returns the footer line count.
  int getFooterLineCount() {
    return this.footerLineCount;
  }

  /// Gets the boolean value indicating the hidden state for the line with given index.
  ///
  /// * index - _required_ - The index of the line for which the hidden state is to be obtained.
  /// * repeatValueCount - _required_ - The number of subsequent lines with same state.
  ///
  /// Returns the boolean value indicating the hidden state for a line.
  List getHidden(int index, int repeatValueCount) {
    return [
      this._lineHidden.getRange(index, repeatValueCount)[0],
      this._lineHidden.getRange(index, repeatValueCount)[1]
    ];
  }

  /// Initializes the scroll axis.
  ///
  /// * scrollAxis - _required_ - The scroll axis.
  void initializeScrollAxis(_ScrollAxisBase scrollAxis) {
    var pixelScrollAxis = scrollAxis as _PixelScrollAxis;
    if (this._lineNested.length > 0 && pixelScrollAxis == null) {
      throw Exception(
          "When you have nested line collections you need to use PixelScrolling!");
    }
    scrollAxis.defaultLineSize = this.defaultLineSize;
    scrollAxis.lineCount = this.lineCount;
    for (int i = 0; i < this._lineSizes.count; i++) {
      if (_lineSizes[i].value != -2) {
        scrollAxis.setLineSize(
            _lineSizes[i].start,
            _lineSizes[i].end,
            _lineSizes[i].value < 0
                ? this.defaultLineSize
                : _lineSizes[i].value);
      }
    }
    for (int i = 0; i < this._lineNested.length; i++)
      this._lineNested.forEach((key, value) {
        pixelScrollAxis.setNestedLines(key, value.distances);
      });
    for (int i = 0; i < this._lineHidden.count; i++) {
      scrollAxis.setLineHiddenState(
          _lineHidden[i].start, _lineHidden[i].end, _lineHidden[i].value);
    }
  }

  /// Gets the distances collection which is used internally for mapping from a point position to
  /// a line index and vice versa.
  ///
  /// Returns the distances collection which is used internally for mapping from a point position
  /// to a line index and vice versa.
  _DistanceCounterCollectionBase get distances {
    if (this._distances == null) {
      this._distances = _DistanceRangeCounterCollection.fromPaddingDistance(
          this.paddingDistance);
      this.initializeDistances();
    }

    return this._distances;
  }

  set distances(_DistanceCounterCollectionBase newValue) {
    if (this._distances == newValue) return;
    this._distances = newValue;
  }

  void initializeDistances() {
    this.distances.clear();
    this.distances.count = getLineCount();
    this.distances.defaultDistance = this.defaultLineSize;
    this._lineNested.forEach((key, value) {
      int repeatSizeCount;
      bool hide = this.getHidden(key, repeatSizeCount)[0];
      repeatSizeCount = this.getHidden(key, repeatSizeCount)[1];
      if (hide) {
        this._distances.setNestedDistances(key, null);
      } else {
        this._distances.setNestedDistances(key, value.distances);
      }
    });

    for (_RangeValuePair<double> entry in this._lineSizes.rangeValues) {
      if (entry.value != -2) {
        this._distances.setRange(entry.start, entry.end,
            entry.value < 0 ? this.defaultLineSize : entry.value);
      }
    }

    for (_RangeValuePair<bool> entry in this._lineHidden.rangeValues) {
      if (entry.value != null && entry.value is bool && entry.value as bool) {
        this.setRange(entry.start, entry.end, 0);
      }
    }
  }

  /// Gets the nested distances, if a line contains a nested lines collection, otherwise null.
  ///
  /// * line - _required_ - The line at which the distances is to be obtained.
  ///
  /// Returns the nested distances, if a line contains a nested lines collection, otherwise null.
  _DistanceCounterCollectionBase getDistances(int line) {
    var nestedLines = this.getNestedLines(line) as _DistancesHostBase;
    if (nestedLines != null) {
      return nestedLines.distances;
    }

    return null;
  }

  /// Inserts lines in the collection and raises the [LinesInserted] event.
  ///
  /// * insertAtLine - _required_ - The index of the first line to insert.
  /// * count - _required_ - The count of the lines to be inserted.
  void insertLinesBase(int insertAtLine, int count) {
    this.insertLines(insertAtLine, count, null);
  }

  /// Insert the given number of lines at the given index.
  ///
  /// * insertAtLine - _required_ - The index of the first line to insert.
  /// * count - _required_ - The count of the lines to be inserted.
  /// * movelines - _required_ - A container with saved state from a preceding Syncfusion.GridCommon.ScrollAxis.LineSizeCollection.InsertLines(int, int)
  /// call when lines should be moved. When it is null, empty lines with default size are inserted.
  @override
  void insertLines(
      int insertAtLine, int count, _EditableLineSizeHostBase movelines) {
    var moveLines = movelines as _LineSizeCollection;
    this._lineSizes.insertWithThreeArgs(
        insertAtLine, count, moveLines == null ? null : moveLines._lineSizes);
    this._lineHidden.insertWithThreeArgs(
        insertAtLine, count, moveLines == null ? null : moveLines._lineHidden);
    this._lineNested = Map<int, _LineSizeCollection>();

    this._lineNested.forEach((key, value) {
      if (key >= insertAtLine) {
        this._lineNested.putIfAbsent(key + count, () => value);
      } else {
        this._lineNested.putIfAbsent(key, () => value);
      }
    });

    if (moveLines != null) {
      for (int i = 0; i < moveLines._lineNested.length; i++)
        moveLines._lineNested.forEach((key, value) {
          this._lineNested.putIfAbsent(key + insertAtLine, () => value);
        });
    }

    this._lineCount += count;

    if (this.isSuspendUpdates) {
      return;
    }

    if (this._distances != null) {
      _DistancesUtil.onInserted(this._distances, this, insertAtLine, count);
    }

    if (this.onLinesInserted != null) {
      var linesInsertedArgs = _LinesInsertedArgs.fromArgs(insertAtLine, count);
      this.onLinesInserted(linesInsertedArgs);
    }
  }

  /// Removes lines from the collection and raises the [LinesRemoved] event.
  ///
  /// * removeAtLine - _required_ - The index of the first line to be removed.
  /// * count - _required_ - The count of the lines to be removed.
  void removeLinesBase(int removeAtLine, int count) {
    this.removeLines(removeAtLine, count, null);
  }

  /// Removes a number of lines at the given index.
  ///
  /// * removeAtLine - _required_ - The index of the first line to be removed.
  /// * count - _required_ - The count of the lines to be removed.
  /// * movelines - _required_ - A container to save state for a subsequent Syncfusion.GridCommon.ScrollAxis.LineSizeCollection.RemoveLines(int, int)
  /// call when lines should be moved.
  @override
  void removeLines(
      int removeAtLine, int count, _EditableLineSizeHostBase movelines) {
    var moveLines = movelines as _LineSizeCollection;
    this._lineSizes.removeWithThreeArgs(
        removeAtLine, count, moveLines == null ? null : moveLines._lineSizes);
    this._lineHidden.removeWithThreeArgs(
        removeAtLine, count, moveLines == null ? null : moveLines._lineHidden);

    Map<int, _LineSizeCollection> lineNested = this._lineNested;
    this._lineNested = Map<int, _LineSizeCollection>();
    for (int i = 0; i < lineNested.length; i++)
      lineNested.forEach((key, value) {
        if (key >= removeAtLine) {
          if (key >= removeAtLine + count) {
            this._lineNested.putIfAbsent(key - count, () => value);
          } else if (moveLines != null) {
            moveLines._lineNested.putIfAbsent(key - removeAtLine, () => value);
          }
        } else {
          this._lineNested.putIfAbsent(key, () => value);
        }
      });

    this._lineCount -= count;

    if (this.isSuspendUpdates) {
      return;
    }

    if (this._distances != null) {
      this._distances.remove(removeAtLine, count);
    }

    if (this.onLinesRemoved != null) {
      var linesRemovedArgs = _LinesRemovedArgs.fromArgs(removeAtLine, count);
      this.onLinesRemoved(linesRemovedArgs);
    }
  }

  /// Gets a value indicating whether the host supports inserting and removing lines or not.
  ///
  /// Returns the boolean value indicating whether the host supports inserting and removing lines.
  bool get supportsInsertRemove => true;

  /// Creates the object which holds temporary state when moving lines.
  ///
  /// Returns the object which holds temporary state when moving lines.
  _EditableLineSizeHostBase createMoveLines() {
    return _LineSizeCollection();
  }

  /// Resets the hidden state of the line.
  void resetHiddenState() {
    this._lineHidden = _SortedRangeValueList<bool>();
  }

  /// Initialize the collection with a pattern of hidden lines.
  ///
  /// * start - _required_ - The index of the first line where the pattern should be
  /// started to be applied.
  /// * lineCount - _required_ - The pattern is applied up to until the lineCount given.
  /// The last initialized line is at index lineCount-1.
  /// * values - _required_ - The pattern that is applied repeatedly.
  void setHiddenInterval(int start, int lineCount, List<bool> values) {
    this.suspendUpdates();

    this._lineHidden = _SortedRangeValueList();

    for (int index = start; index < lineCount; index += values.length) {
      for (int n = 0; n < values.length; n++) {
        if (n + index < lineCount) {
          this._lineHidden[index + n] = values[n];
        }
      }
    }

    this.resumeUpdates();
  }

  /// Initialize the collection with a pattern of hidden lines with state.
  ///
  /// * start - _required_ - The index of the first line where the pattern should be
  /// started to be applied.
  /// * lineCount - _required_ - The pattern is applied up to until the lineCount given.
  /// The last initialized line is at index lineCount-1.
  /// * values - _required_ - The pattern that is applied repeatedly.
  void setHiddenIntervalWithState(int start, int lineCount, List<bool> values) {
    this.suspendUpdates();

    for (int index = start; index < lineCount; index += values.length) {
      for (int n = 0; n < values.length; n++) {
        if (n + index < lineCount) {
          this._lineHidden[index + n] = values[n];
        }
      }
    }

    this.resumeUpdates();
  }

  /// Set the hidden state all at once in one operation.
  ///
  /// Use this method if you want to change the hidden
  /// state of many rows at once since this will be much faster instead of individually setting rows hidden.
  ///
  /// * values - _required_ - The new hidden state for rows.
  void setHiddenState(List<bool> values) {
    this.suspendUpdates();

    this._lineHidden = _SortedRangeValueList<bool>();

    int count = min(this.lineCount, values.length);
    for (int index = 0; index < count; index++) {
      this._lineHidden[index] = values[index];
    }

    this.resumeUpdates();
  }

  int _isSuspendUpdates = 0;
  bool get isSuspendUpdates => this._isSuspendUpdates > 0;

  /// Suspends the updates in the view.
  void suspendUpdates() {
    this._isSuspendUpdates++;
  }

  /// Resumes the update in the view.
  void resumeUpdates() {
    this._isSuspendUpdates--;
    if (this._isSuspendUpdates == 0) {
      if (this._distances != null) {
        this.initializeDistances();
      }

      if (this.onLineHiddenChanged != null) {
        var hiddenRangeChangedArgs =
            _HiddenRangeChangedArgs.fromArgs(0, this._lineCount - 1, false);
        this.onLineHiddenChanged(hiddenRangeChangedArgs);
      }
    }
  }
}

/// Initializes a new instance of the DistanceLineCounterTree class.
class _DistancesUtil {
  /// Prevents a default instance of the DistancesUtil class from being created.
  _DistancesUtil();

  /// Gets the range to helper.
  ///
  /// * n - _required_ - The index.
  /// * to - _required_ - The end index.
  /// * repeatSizeCount - _required_ - The count of repeated sizes.
  ///
  /// Returns the minimum index value
  static int getRangeToHelper(int n, int to, int repeatSizeCount) {
    if (repeatSizeCount == _MathHelper.maxvalue) {
      return to;
    }

    return min(to, n + repeatSizeCount - 1);
  }

  /// This method fires when the lines is inserted.
  ///
  /// * distances - _required_ - The distances
  /// * linesHost - _required_ - The line host.
  /// * insertAt - _required_ - The index to insert.
  /// * count - _required_ - The count of the lines inserted.
  static void onInserted(_DistanceCounterCollectionBase distances,
      _LineSizeHostBase linesHost, int insertAt, int count) {
    distances.insert(insertAt, count);
    int to = insertAt + count - 1;
    int repeatSizeCount;

    // Set line sizes
    for (int index = insertAt; index <= to; index++) {
      double size = linesHost.getSize(index, repeatSizeCount)[0];
      repeatSizeCount = linesHost.getSize(index, repeatSizeCount)[1];
      if (size != distances.defaultDistance) {
        int rangeTo = getRangeToHelper(index, to, repeatSizeCount);
        distances.setRange(index, rangeTo, size);
        index = rangeTo;
      }
    }

    // Also check for hidden rows and reset line sizes for them.
    for (int index = insertAt; index <= to; index++) {
      bool hide = linesHost.getHidden(index, repeatSizeCount)[0];
      if (hide) {
        int rangeTo = getRangeToHelper(index, to, repeatSizeCount);
        distances.setRange(index, rangeTo, 0.0);
        index = rangeTo;
      }
    }
  }

  /// This method fires when distances line hidden changed.
  ///
  /// * distances - _required_ - The distances
  /// * linesHost - _required_ - The line host.
  /// * from - _required_ - The start index of the line.
  /// * to - _required_ - The end index of the line.
  static void distancesLineHiddenChanged(
      _DistanceCounterCollectionBase distances,
      _LineSizeHostBase linesHost,
      int from,
      int to) {
    var ndh = linesHost as _NestedDistancesHostBase;
    for (int n = from; n <= to; n++) {
      int repeatSizeCount;
      bool hide = linesHost.getHidden(n, repeatSizeCount)[0];
      repeatSizeCount = linesHost.getHidden(n, repeatSizeCount)[1];
      if (ndh == null || ndh.getDistances(n) == null) {
        int rangeTo = getRangeToHelper(n, to, repeatSizeCount);
        if (hide) {
          distances.setRange(n, rangeTo, 0.0);
        } else {
          distancesLineSizeChanged(distances, linesHost, n, rangeTo);
        }

        n = rangeTo;
      } else {
        distances.setNestedDistances(n, hide ? null : ndh.getDistances(n));
      }
    }
  }

  /// This method fires when the distances line size changed.
  ///
  /// * distances - _required_ - The distances
  /// * linesHost - _required_ - The line host.
  /// * from - _required_ - The start index of the line.
  /// * to - _required_ - The end index of the line.
  static void distancesLineSizeChanged(_DistanceCounterCollectionBase distances,
      _LineSizeHostBase linesHost, int from, int to) {
    var ndh = linesHost as _NestedDistancesHostBase;
    for (int n = from; n <= to; n++) {
      if (ndh == null || ndh.getDistances(n) == null) {
        int repeatSizeCount;
        double size = linesHost.getSize(n, repeatSizeCount)[0];
        repeatSizeCount = linesHost.getSize(n, repeatSizeCount)[1];
        int rangeTo = getRangeToHelper(n, to, repeatSizeCount);
        distances.setRange(n, rangeTo, size);
        n = rangeTo;
      } else {
        distances.setNestedDistances(n, ndh.getDistances(n));
      }
    }
  }
}
