part of datagrid;

/// Returns the PropertyChangedArgs by used the [onPropertyChanged] event.
typedef _PropertyChangedCallback = void Function(
    _PropertyChangedArgs propertyChangedArgs);

/// Returns the valueChangeArgs by used the [onValueChanged] event.
typedef _ValueChangedCallback = void Function();

/// Returns the ValueChangingArgs by used the [onValueChanging] event.
typedef _ValueChangingCallback = void Function(
    _ValueChangingArgs valueChangingArgs);

/// Provides all properties to configure a scrollbar.
class _ScrollInfo extends _ScrollBarBase {
  double _value = 0;
  double _minimum = 0;
  double _maximum = 100;
  double _largeChange = 10;
  double _smallChange = 1;
  double _proposedLargeChange = 10;
  bool _enabled = true;

  /// Gets a value indicating whether the scroll bar is enabled or not.
  ///
  /// Returns a boolean value that indicates whether the scroll bar is enabled.
  @override
  bool get enabled => this._enabled;

  /// Sets a value indicating whether the scroll bar is enabled or not.
  set enabled(bool value) {
    if (this._enabled != value) {
      this._enabled = value;
    }
  }

  /// Gets a value to be added to or subtracted from the value of the property when the scroll box is moved a large distance.
  ///
  /// Returns a value to be added to or subtracted from the value of the property when the scroll box is moved a large distance.
  @override
  double get largeChange => this._largeChange;

  /// Sets a value to be added to or subtracted from the value of the property when the scroll box is moved a large distance.
  set largeChange(double value) {
    this._proposedLargeChange = value;
    if (value < 0) {
      value = this.maximum - this.minimum;
    }

    if (this._largeChange != value) {
      this._largeChange = value;
    }
  }

  /// Gets the upper limit of values of the scrollable range.
  ///
  /// Returns the upper limit of values of the scrollable range.
  double get maximum => this._maximum;

  /// Sets the upper limit of values of the scrollable range.
  set maximum(double value) {
    if (this._maximum != value) {
      this._maximum = value;
    }
  }

  /// Gets the lower limit of values of the scrollable range.
  ///
  /// Returns the lower limit of values of the scrollable range.
  double get minimum => this._minimum;

  /// Sets the lower limit of values of the scrollable range.
  set minimum(double value) {
    if (this._minimum != value) {
      this._minimum = value;
    }
  }

  /// Gets the value to be added to or subtracted from the value of the property when the scroll box is moved a small distance.
  ///
  /// Retunrs the value to be added to or subtracted from the value of the property when the scroll box is moved a small distance.
  double get smallChange => this._smallChange;

  /// Sets the value to be added to or subtracted from the value of the property when the scroll box is moved a small distance.
  set smallChange(double value) {
    if (this._smallChange != value) {
      this._smallChange = value;
    }
  }

  /// Gets a numeric value that represents the current position of the scroll box on the scroll bar control.
  ///
  /// Returns a numeric value that represents the current position of the scroll box on the scroll bar control.
  double get value {
    if (this._proposedLargeChange < 0) {
      return this._minimum;
    }
    return max(
        this._minimum, min(this._maximum - this._largeChange + 1, this._value));
  }

  /// Sets a numeric value that represents the current position of the scroll box on the scroll bar control.
  set value(double value) {
    if (this.value != value) {
      var e = _ValueChangingArgs(value, this.value);
      if (this.onValueChanging != null) {
        this.onValueChanging(e);
      }

      if (!e.cancel) {
        double offset = e.newValue;

        if (offset < this.minimum ||
            this.largeChange >= this.maximum - this.minimum) {
          offset = this.minimum;
        } else {
          if (offset + this.largeChange > this.maximum) {
            offset = max(this.minimum, this.maximum - this.largeChange);
          }
        }

        this._value = offset;
      }
    }
  }

  ///Initializes a new instance of the [ScrollInfo] class.
  _ScrollInfo();

  /// Called when a property is changed and raises the [PropertyChanged] event.
  ///
  /// * propertyName - _required_ - Name of the property.
  void onPropertyChanged(String propertyName) {
    if (this.onPropertyChangedEvent != null) {
      var propertyChangedArgs = _PropertyChangedArgs(propertyName);
      this.onPropertyChangedEvent(propertyChangedArgs);
    }
  }

  /// Occurs when a property value changes.
  _PropertyChangedCallback onPropertyChangedEvent;

  /// Occurs when the current position of the scroll box on the scroll bar has changed.
  _ValueChangedCallback onValueChanged;

  /// Occurs when the current position of the scroll box on the scroll bar is being changed.
  _ValueChangingCallback onValueChanging;

  /// Copies current settings to another object.
  ///
  /// * scrollBar - _required_ - The object to which the settings is to be copied.
  void copyTo(_ScrollInfo scrollBar) {
    scrollBar.value = this._value;
    scrollBar.minimum = this._minimum;
    scrollBar.maximum = this._maximum;
    scrollBar.largeChange = this._largeChange;
    scrollBar.smallChange = this._smallChange;
    scrollBar.enabled = this._enabled;
  }

  List list;

  /// Clones this instance.
  ///
  /// Returns the cloned instance.
  _ScrollInfo clone() {
    var sb = _ScrollInfo();
    this.copyTo(sb);
    return sb;
  }

  /// Determines whether the specified `ScrollInfo` is equal to the current `ScrollInfo`.
  ///
  /// * obj - _required_ - The `ScrollInfo` to compare with the current `ScrollInfo`.
  ///
  /// Returns `True` if the specified `ScrollInfo` is equal to the current `ScrollInfo`, otherwise `false`.
  bool equals(Object obj) {
    var sb = obj as _ScrollInfo;
    if (sb == null && this == null) {
      return true;
    } else if (this == null || sb == null) {
      return false;
    }

    return sb.value == this._value &&
        sb.minimum == this._minimum &&
        sb.maximum == this._maximum &&
        sb.largeChange == this._largeChange &&
        sb.smallChange == this._smallChange &&
        sb.enabled == this._enabled;
  }

  /// Serves as a hash function for a particular type.
  ///
  /// returns a hash code for the current `ScrollInfo`.
  int getHashCode() {
    return this.value.hashCode;
  }

  /// A `string` that represents the current `object`.
  ///
  /// Returns a `string` that represents the current `object`.
  @override
  String toString() {
    return "ScrollInfo ( Value = ${this.value}, Minimum = ${this.minimum}, Maximum = ${this.maximum}, LargeChange = ${this.largeChange}, Enabled = ${this.enabled} )";
  }
}

/// Provides data for the `ScrollInfo.ValueChanging` event.
class _ValueChangingArgs {
  double _newValue = 0.0;
  double _oldValue = 0.0;

  /// Initializes a new instance of the ValueChangingArgs class.
  ///
  /// * newValue - _required_ - The new value.
  /// * oldValue - _required_ - The old value.
  _ValueChangingArgs(double newValue, double oldValue) {
    this._newValue = newValue;
    this._oldValue = oldValue;
  }

  /// Gets a value indicating whether to cancel the value change in scroll bar.
  ///
  /// Returns a boolean value indicating whether to cancel the value change in scroll bar.
  bool get cancel => _cancel;
  bool _cancel = false;

  /// Sets a value indicating whether to cancel the value change in scroll bar.
  set cancel(bool value) {
    if (value == _cancel) return;
    _cancel = value;
  }

  /// Gets the new value.
  ///
  /// Returns the new value.
  double get newValue => this._newValue;

  /// Gets the old value.
  ///
  /// Returns the old value.
  double get oldValue => this._oldValue;
}

class _PropertyChangedArgs {
  ///  Occurs when name of the property was changed.
  _PropertyChangedArgs(String propertyName);

  /// Gets the name of the property that changed.
  ///
  /// Returns the property name.
  String get propertyName => _propertyName;
  String _propertyName;
}
