part of datagrid;

/// Holds a range together with a value assigned to the range.
///
/// * T - _required_ - The type of the parameter.
class _RangeValuePair<T> extends Comparable<T> {
  int _start;
  int _count;
  dynamic _value;

  /// Initializes a new instance of the _RangeValuePair class.
  ///
  /// * start - _required_ - The start of the range.
  _RangeValuePair(int start) {
    this._start = start;
    this._count = 1;
    this._value = 0;
  }

  /// Initializes a new instance of the _RangeValuePair class.
  ///
  /// * start - _required_ - The start of the range.
  /// * count - _required_ - The count of the range.
  /// * value - _required_ - The value for the range.
  _RangeValuePair.fromRangeValuePair(int start, int count, dynamic value) {
    this._start = start;
    this._count = count;
    this._value = value;
  }

  /// Gets the start  of the range.
  ///
  /// Returns the start of the range.
  int get start => this._start;

  /// Sets the start  of the range.
  set start(int value) {
    this._start = value;
  }

  /// Gets the count of the range.
  ///
  /// Returns the count of the range.
  int get count => this._count;

  /// Sets the count of the range.
  set count(int value) {
    this._count = value;
  }

  /// Gets the value for the range.
  ///
  /// Returns the value for the range.
  dynamic get value => this._value;

  /// Sets the value for the range.
  set value(dynamic value) {
    this._value = value;
  }

  /// Gets the end of the range.
  ///
  /// Returns the end of the range.
  int get end => this._start + this._count - 1;

  /// Sets the end of the range.
  set end(int value) {
    this._count = value - this._start + 1;
  }

  /// Compares the current range with the range of the other object.
  ///
  /// A 32-bit signed integer that indicates the relative order of the objects being compared.
  ///
  /// * obj - _required_ - An object to compare with this instance.

  int compareTo(Object obj) {
    _RangeValuePair<T> x = this;
    var y = obj as _RangeValuePair<T>;

    if (((x._start >= y._start) && (x._start < y._start + y._count)) ||
        ((y._start >= x._start) && (y._start < x._start + x._count))) {
      return 0;
    }

    return this._start.compareTo(y._start);
  }

  /// Gets the Debug / text information about the node.
  ///
  /// Returns a `string` with state information about this `object`.
  @override
  String toString() {
    return "$runtimeType Start = ${this._start} Count = ${this._count} End = ${this.end} Value = ${this._value}";
  }
}

/// Maintain a collection of _RangeValuePair
///
/// A sorted list with `_RangeValuePair{T}` ordered by the
/// start index of the ranges. _SortedRangeValueList ensures that ranges
/// of the elements inside the list do not overlap and it also ensures
/// that there are no empty gaps meaning that the subsequent range will
/// always have the Start position be set to the End position of the previous
/// range plus one.
///
/// * T - _required_ - The type of the parameter.
class _SortedRangeValueList<T>
    extends _EnumerableGenericBase<_RangeValuePair<T>> {
  List<_RangeValuePair<T>> rangeValues = List<_RangeValuePair<T>>();
  T _defaultValue;

  /// Initializes a new instance of the _SortedRangeValueList class.
  _SortedRangeValueList();

  /// Initializes a new instance of the _SortedRangeValueList class.
  ///
  /// * defaultValue - _required_ - The default value used for filling gaps.
  _SortedRangeValueList.from(T defaultValue) {
    this._defaultValue = defaultValue;
  }

  /// Gets the default value used for filling gaps.
  ///
  /// Returns the default value used for filling gaps.
  dynamic get defaultValue {
    if (this._defaultValue == null) {
      if (_defaultValue.runtimeType is bool) {
        return false;
      }
      if (_defaultValue.runtimeType is double) {
        return 0.0;
      }
    } else {
      return this._defaultValue;
    }
  }

  /// Sets the default value used for filling gaps.
  set defaultValue(T value) {
    this._defaultValue = value;
  }

  /// Clears the stored ranges.
  void clear() {
    this.rangeValues.clear();
  }

  /// Gets the count which is the same as the end position of the last range.
  ///
  /// Returns the count which is the same as the end position of the last range.
  int get count {
    if (this.rangeValues.length == 0) {
      return 0;
    }
    _RangeValuePair<T> rv = this.rangeValues[this.rangeValues.length - 1];
    return rv.start + rv.count;
  }

  _RangeValuePair<T> getRangeValue(int index) {
    int n = _MathHelper.binarySearch<_RangeValuePair<T>>(
        this.rangeValues, _RangeValuePair<T>(index));
    return this.rangeValues[n];
  }

  /// Gets the value of the range that contains the specified index
  ///
  /// When necessary it splits a range and creates
  /// a new range value pair to hold the new value for the specified index.
  ///
  /// index - _required_ - index value
  /// Returns the value of the range that contains the specified index.
  ///
  /// Returns the value range for the specified index.
  operator [](index) {
    return this.getRangeValue(index).value;
  }

  operator []=(index, value) {
    bool b = false;
    if (index >= this.count) {
      if (value == (this.defaultValue)) {
        return;
      }

      b = true;
    }

    if (b || (value != this[index])) {
      this.ensureCount(index);
      int n = this.split(index);
      this.split(index + 1);
      if (n == this.rangeValues.length) {
        if (n > 0 && (value == this.rangeValues[n - 1].value)) {
          this.rangeValues[n - 1].count++;
        } else {
          this
              .rangeValues
              .add(_RangeValuePair<T>.fromRangeValuePair(index, 1, value));
        }
      } else {
        this.rangeValues[n].value = value;
      }
    }
  }

  /// Gets a range that contains the specified index and also returns a count indicating the delta between the index and the
  /// end range.
  ///
  /// * index - _required_ - The index of the range.
  /// * count - _required_ - The count of the range.
  ///
  /// Returns a count indicating the delta
  List getRange(dynamic index, dynamic count) {
    if (index >= this.count) {
      count = _MathHelper.maxvalue;
      return [this.defaultValue, count];
    }

    _RangeValuePair<T> rv = this.getRangeValue(index);
    count = rv.end - index + 1;
    return [rv.value, count];
  }

  /// Inserts a range initialized with [DefaultValue] at
  /// the specified index. When necessary it splits a range and creates
  /// a new range value pair.
  ///
  /// * insertAt - _required_ - The insertion point.
  /// * count - _required_ - The count of the items to be inserted.
  void insert(dynamic insertAt, dynamic count) {
    this.insertWithFourArgs(insertAt, count, this.defaultValue, null);
  }

  /// Inserts a range initialized with a given value at
  /// the specified index.
  ///
  /// When necessary it splits a range and creates
  /// a new range value pair.
  ///
  /// * insertAt - _required_ - The insertion point.
  /// * count - _required_ - The count.
  /// * value - _required_ - The value.
  void insertwithThreeArgsGeneric(
      dynamic insertAt, dynamic count, dynamic value) {
    this.insertWithFourArgs(insertAt, count, value, null);
  }

  /// Inserts a range initialized with [DefaultValue] at
  /// the specified index.
  ///
  /// When necessary it splits a range and creates
  /// a new range value pair.
  ///
  /// * insertAt - _required_ - The insertion point.
  /// * count - _required_ - The count.
  /// * moveRanges - _required_ - Allocate this object before a preceding remove call when moving ranges.
  /// Otherwise specify null.
  void insertWithThreeArgs(
      dynamic insertAt, dynamic count, _SortedRangeValueList<T> moveRanges) {
    this.insertWithFourArgs(insertAt, count, this.defaultValue, moveRanges);
  }

  /// Inserts a range initialized with a given value at
  /// the specified index.
  ///
  /// When necessary it splits a range and creates
  /// a new range value pair.
  ///
  /// * insertAt - _required_ - The insertion point.
  /// * count - _required_ - The count.
  /// * value - _required_ - The value.
  /// * moveRanges - _required_ - Allocate this object before a preceding Remove call when moving ranges.
  /// Otherwise specify null.
  void insertWithFourArgs(dynamic insertAt, dynamic count, dynamic value,
      _SortedRangeValueList<T> moveRanges) {
    if (insertAt >= this.count) {
      if (value == (this.defaultValue) &&
          (moveRanges == null || moveRanges.count == 0)) {
        return;
      }

      this.ensureCount(insertAt);
      this
          .rangeValues
          .add(_RangeValuePair<T>.fromRangeValuePair(insertAt, count, value));
      if (this.rangeValues.length >= 2) {
        this.merge(this.rangeValues.length - 2);
      }
    } else {
      dynamic n = _MathHelper.binarySearch<_RangeValuePair<T>>(
          this.rangeValues, _RangeValuePair<T>(insertAt));
      _RangeValuePair<T> rv = this.rangeValues[n];
      if (value == (rv.value)) {
        rv.count += count;
        this.adjustStart(n + 1, count);
      } else {
        n = this.splitWithTwoArgs(insertAt, n);
        split(insertAt + 1);
        var rv2 = _RangeValuePair<T>.fromRangeValuePair(insertAt, count, value);
        this.rangeValues.insert(n, rv2);
        adjustStart(n + 1, count);
        this.merge(n);
        if (n > 0) {
          this.merge(n - 1);
        }
      }
    }

    if (moveRanges != null) {
      for (int i = 0; i < moveRanges.count; i++) {
        setRange(moveRanges[i].start + insertAt, moveRanges[i].count,
            moveRanges[i].value);
      }
    }
  }

  /// Removes a range at the specified index.
  ///
  /// When necessary ranges
  /// are merged when preceding and subsequent ranges have the same
  /// value.
  ///
  /// * removeAt - _required_ - The index for the range to be removed.
  /// * count - _required_ - The count.
  void remove(dynamic removeAt, dynamic count) {
    this.removeWithThreeArgs(removeAt, count, null);
  }

  /// Removes a range at the specified index.
  ///
  /// When necessary ranges
  /// are merged when preceding and subsequent ranges have the same
  /// value.
  ///
  /// * removeAt - _required_ - The index for the range to be removed.
  /// * count - _required_ - The count.
  /// * moveRanges - _required_ - Allocate this object before a remove call when moving ranges
  /// and pass it to a subsequent insert call. Otherwise specify null.
  void removeWithThreeArgs(
      dynamic removeAt, dynamic count, _SortedRangeValueList<T> moveRanges) {
    if (removeAt >= this.count) {
      return;
    }

    dynamic n = this.removeHelper(removeAt, count, moveRanges);
    this.adjustStart(n, -count);
    if (n > 0) {
      this.merge(n - 1);
    }
  }

  /// Sets the value for a range at the specified index.
  ///
  /// When necessary ranges
  /// are split or merged to make sure integrity of the list is maintained.
  /// (_SortedRangeValueList ensures that ranges
  /// of the elements inside the list do not overlap and it also ensures
  /// that there are no empty gaps meaning that the subsequent range will
  /// always have the Start position be set to the End position of the previous
  /// range plus one.)
  ///
  /// * index - _required_ - The index for the range to be changed.
  /// * count - _required_ - The count.
  /// * value - _required_ - The value.
  void setRange(dynamic index, dynamic count, dynamic value) {
    if (index >= this.count && value == (this.defaultValue)) {
      return;
    }

    this.ensureCount(index);
    dynamic n = this.removeHelper(index, count, null);
    var rv = _RangeValuePair<T>.fromRangeValuePair(index, count, value);
    this.rangeValues.insert(n, rv);
    this.merge(n);
    if (n > 0) {
      this.merge(n - 1);
    }
  }

  void ensureCount(dynamic index) {
    if (index - this.count > 0) {
      this.rangeValues.add(_RangeValuePair<T>.fromRangeValuePair(
          this.count, index - this.count, this.defaultValue));
    }
  }

  void adjustStart(dynamic n, dynamic delta) {
    while (n < this.rangeValues.length) {
      this.rangeValues[n++].start += delta;
    }
  }

  dynamic removeHelper(
      dynamic removeAt, dynamic count, _SortedRangeValueList<T> moveRanges) {
    if (removeAt >= this.count) {
      return this.rangeValues.length;
    }

    dynamic n = split(removeAt);
    split(removeAt + count);
    dynamic total = 0;
    dynamic deleteCount = 0;
    while (total < count && n + deleteCount < this.rangeValues.length) {
      _RangeValuePair<T> rv = this.rangeValues[n + deleteCount];
      total += rv.count;
      deleteCount++;
      if (moveRanges != null && !(rv.value == this.defaultValue)) {
        moveRanges.rangeValues.add(_RangeValuePair<T>.fromRangeValuePair(
            rv.start - removeAt, rv.count, rv.value));
      }
    }

    this.rangeValues.removeRange(n, (n + deleteCount));
    return n;
  }

  dynamic split(dynamic index) {
    if (index >= this.count) {
      return this.rangeValues.length;
    }
    int n = _MathHelper.binarySearch<_RangeValuePair<T>>(
        this.rangeValues, _RangeValuePair<T>(index));
    return this.splitWithTwoArgs(index, n);
  }

  int splitWithTwoArgs(dynamic index, int n) {
    _RangeValuePair<T> rv = this.rangeValues[n];
    if (this.rangeValues[n].start == index) {
      return n;
    }

    int count1 = index - this.rangeValues[n].start;
    int count2 = this.rangeValues[n].count - count1;
    rv.count = count1;

    var rv2 = _RangeValuePair<T>.fromRangeValuePair(index, count2, rv.value);
    this.rangeValues.insert(n + 1, rv2);
    return n + 1;
  }

  void merge(int n) {
    if (n >= this.rangeValues.length) {
      return;
    }

    _RangeValuePair<T> rv1 = this.rangeValues[n];
    if (n == this.rangeValues.length - 1) {
      if (rv1.value == this.defaultValue) {
        this.rangeValues.removeAt(n);
      }

      return;
    }

    _RangeValuePair<T> rv2 = this.rangeValues[n + 1];
    if (rv1.value == rv2.value) {
      rv1.count += rv2.count;
      this.rangeValues.removeAt(n + 1);
    }
  }

  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}
