part of datagrid;

typedef _DataGridStateDetails = _DataGridSettings Function();

typedef QueryCellStyleCallback = DataGridCellStyle Function(
    QueryCellStyleArgs queryCellStyleArgs);

typedef QueryRowStyleCallback = DataGridCellStyle Function(
    QueryRowStyleArgs queryRowStyleArgs);

typedef CellBuilderCallback = Widget Function(
    BuildContext contex, GridColumn column, int rowIndex);

typedef QueryRowHeightCallback = double Function(int rowIndex);

typedef SelectionChangingCallback = bool Function(
    List<Object> addedRows, List<Object> removedRows);

typedef SelectionChangedCallback = void Function(
    List<Object> addedRows, List<Object> removedRows);

typedef CellRenderersCreatedCallback = void Function(
    Map<String, GridCellRendererBase> cellRenderers);

typedef CurrentCellActivatingCallback = bool Function(
    RowColumnIndex newRowColumnIndex, RowColumnIndex oldRowColumnIndex);

typedef CurrentCellActivatedCallback = void Function(
    RowColumnIndex newRowColumnIndex, RowColumnIndex oldRowColumnIndex);

/// A material design datagrid.
///
/// DataGrid lets you display and manipulate data in a tabular view. It is built from the ground up to achieve the best possible performance even when loading large amounts of data.
///
/// DataGrid supports different types of column types to populate the columns for different types of data such as int, double, DateTime, String.
///
/// You can use [GridWidgetColumn] to load any widget in a column. [source] property enables you to populate the data for the [SfDataGrid].
///
/// This sample shows how to populate the data for the [SfDataGrid] and display with four columns: id, name, designation and salary.
/// The columns are defined by four [GridColumn] objects.
///
/// ``` dart
///   final List<Employee> _employees = <Employee>[];
///   final EmployeeDataSource _employeeDataSource = EmployeeDataSource();
///
///   @override
///   void initState()
///   {
///     super.initState();
///     populateData();
///   }
///
///   @override
///   Widget build(BuildContext context) {
///     return SfDataGrid(
///       source: _employeeDataSource,
///       columns: [
///       GridNumericColumn(mappingName: 'id')
///       ..headerText = 'ID',
///       GridTextColumn(mappingName: 'name')
///       ..headerText = 'Name',
///       GridTextColumn(mappingName: 'designation')
///       ..headerText = 'Designation',
///       GridNumericColumn(mappingName: 'salary')
///       ..headerText = 'Salary',
///     ],
///     );
///   }
///
///   void populateData()
///   {
///     _employees.add(Employee(10001, 'James', 'Project Lead', 10000));
///     _employees.add(Employee(10002, 'Kathryn', 'Manager', 10000));
///     _employees.add(Employee(10003, 'Lara', 'Developer', 10000));
///     _employees.add(Employee(10004, 'Michael', 'Designer', 10000));
///     _employees.add(Employee(10005, 'Martin', 'Developer', 10000));
///     _employees.add(Employee(10006, 'Newberry', 'Developer', 15000));
///     _employees.add(Employee(10007, 'Balnc', 'Developer', 15000));
/// 	  _employees.add(Employee(10008, 'Perry', 'Developer', 15000));
///     _employees.add(Employee(10009, 'Gable', 'Developer', 15000));
///     _employees.add(Employee(10010, 'Grimes', 'Developer', 15000));
///
///   }
/// }
///
/// class Employee {
///   Employee(this.id, this.name, this.designation, this.salary);
///   final int id;
///   final String name;
///   final String designation;
///   final int salary;
/// }
///
/// class EmployeeDataSource extends DataGridSource
/// {
///   @override
///   List<Object> get dataSource => _employees;
///
///   @override
///   getCellValue(int rowIndex, String columnName){
///     switch (columnName) {
///       case 'id':
///         return _employees[rowIndex].id;
///         break;
///       case 'name':
///         return _employees[rowIndex].name;
///         break;
///       case 'salary':
///         return _employees[rowIndex].salary;
///         break;
///       case 'designation':
///         return _employees[rowIndex].designation;
///         break;
///       default:
///         return ‘ ’;
///         break;
///    }
///   }
///
/// ```
class SfDataGrid extends StatefulWidget {
  const SfDataGrid(
      {Key key,
      @required this.source,
      @required this.columns,
      double rowHeight,
      double headerRowHeight,
      double defaultColumnWidth,
      GridLinesVisibility gridLinesVisibility,
      ColumnWidthCalculationMode columnWidthCalculationMode,
      ColumnWidthCalculationRange columnWidthCalculationRange,
      ColumnWidthMode columnWidthMode,
      SelectionMode selectionMode,
      GridNavigationMode navigationMode,
      this.selectionManager,
      this.controller,
      this.columnSizer,
      this.cellBuilder,
      this.onQueryCellStyle,
      this.onQueryRowStyle,
      this.onQueryRowHeight,
      this.onSelectionChanged,
      this.onSelectionChanging,
      this.onCellRenderersCreated,
      this.onCurrentCellActivating,
      this.onCurrentCellActivated})
      : assert(source != null),
        assert(columns != null),
        this.rowHeight = rowHeight ?? 48.0,
        this.headerRowHeight = headerRowHeight ?? 56.0,
        this.defaultColumnWidth = defaultColumnWidth ?? (kIsWeb ? 100.0 : 90.0),
        this.gridLinesVisibility =
            gridLinesVisibility ?? GridLinesVisibility.horizontal,
        this.columnWidthCalculationMode =
            columnWidthCalculationMode ?? ColumnWidthCalculationMode.textSize,
        this.columnWidthMode = columnWidthMode ?? ColumnWidthMode.none,
        this.columnWidthCalculationRange = columnWidthCalculationRange ??
            ColumnWidthCalculationRange.visibleRows,
        this.selectionMode = selectionMode ?? SelectionMode.none,
        this.navigationMode = navigationMode ?? GridNavigationMode.row,
        super(key: key);

  /// The height of each row except the column header.
  ///
  /// Defaults to 48.0
  final double rowHeight;

  /// The height of the column header row.
  ///
  ///Defaults to 56.0
  final double headerRowHeight;

  /// The collection of the [GridColumn].
  ///
  /// Each column associated with its own renderer and it controls the corresponding column related operations.
  ///
  /// Defaults to null.
  final List<GridColumn> columns;

  /// The datasource that provides the data for each row in [SfDataGrid]. Must be non-null.
  ///
  /// This object is expected to be long-lived, not recreated with each build.
  ///
  /// Defaults to null
  final DataGridSource source;

  /// The width of each column.
  ///
  /// If the [columnWidthMode] is set for [SfDataGrid] or [GridColumn], or [GridColumn.width] is set, [defaultColumnWidth] will not be considered.
  ///
  /// Defaults to 90.0 for Android & iOS and 100.0 for Web.
  final double defaultColumnWidth;

  /// How the column widths are determined.
  ///
  /// Defaults to [ColumnWidthMode.none]
  ///
  /// Also refer [ColumnWidthMode]
  final ColumnWidthMode columnWidthMode;

  /// How the column widths should be calculated.
  ///
  /// Provides options to calculate whether each cell in a column should be measured based on the size of the text or should be measured based on the length of the text.
  /// When the [ColumnWidthCalculationMode.textLength], the text in each cell which has large length is considered for text size measurement.
  ///
  /// Defaults to [ColumnWidthCalculationMode.textSize]
  ///
  /// Also refer [ColumnWidthCalculationMode]
  final ColumnWidthCalculationMode columnWidthCalculationMode;

  /// The [ColumnSizer] used to control the column width sizing operation of each columns.
  ///
  /// You can override the available methods and customize the required operations in the custom [ColumnSizer].
  final ColumnSizer columnSizer;

  /// How the row count should be considered when calculating the width of a column.
  ///
  /// Provides options to consider only visible rows or all the rows which are available in [SfDataGrid].
  ///
  /// Defaults to [ColumnWidthCalculationRange.visibleRows]
  ///
  /// Also refer [ColumnWidthCalculationRange]
  final ColumnWidthCalculationRange columnWidthCalculationRange;

  /// How the border should be visible.
  ///
  /// Decides whether vertical, horizontal, both the borders and no borders should be drawn.
  ///
  /// Defaults to [GridLinesVisibility.horizontal]
  ///
  /// Also refer [GridLinesVisibility]
  final GridLinesVisibility gridLinesVisibility;

  /// Invoked when the style for each cell is applied.
  ///
  /// Users can set the styling for the cells based on the condition.
  final QueryCellStyleCallback onQueryCellStyle;

  /// Invoked when the style for each row is applied.
  ///
  /// Users can set the styling for the cells based on the condition.
  final QueryRowStyleCallback onQueryRowStyle;

  /// A builder that sets the widget for the [GridWidgetColumn].
  ///
  /// The widget returned by this method is wrapped in a cell.
  final CellBuilderCallback cellBuilder;

  /// Invoked when the row height for each row is queried.
  final QueryRowHeightCallback onQueryRowHeight;

  /// How the rows should be selected.
  ///
  /// Provides options to select single row or multiple rows.
  ///
  /// Defaults to [SelectionMode.none].
  ///
  /// Also refer [SelectionMode]
  final SelectionMode selectionMode;

  /// Invoked when the row is selected.
  ///
  /// This callback never be called when the [onSelectionChanging] is returned as [false].
  final SelectionChangedCallback onSelectionChanged;

  /// Invoked when the row is being selected or being unselected
  ///
  /// This callback's return type is [bool]. So, if you want to cancel the selection on a row based on the condition, return [false]. Otherwise, return [true].
  final SelectionChangingCallback onSelectionChanging;

  /// The [SelectionManagerBase] used to control the selection operations in [SfDataGrid].
  ///
  /// You can override the available methods and customize the required operations in the custom [RowSelectionManager].
  ///
  /// Defaults to null
  final SelectionManagerBase selectionManager;

  /// The [DataGridController] used to control the current cell navigation and selection operation.
  ///
  /// Defaults to null.
  ///
  /// This object is expected to be long-lived, not recreated with each build.
  final DataGridController controller;

  /// Called when the cell renderers are created for each column.
  ///
  /// This method is called once when the [SfDataGrid] is loaded. Users can provide the custom cell renderer to the existing collection.
  final CellRenderersCreatedCallback onCellRenderersCreated;

  final GridNavigationMode navigationMode;

  /// Invoked when the cell is activated.
  ///
  /// This callback never be called when the [onCurrentCellActivating] is returned as [false].
  final CurrentCellActivatedCallback onCurrentCellActivated;

  /// Invoked when the cell is being activated.
  ///
  /// This callback's return type is [bool]. So, if you want to cancel cell activation based on the condition, return [false]. Otherwise, return [true].
  final CurrentCellActivatingCallback onCurrentCellActivating;

  @override
  State<StatefulWidget> createState() => _SfDataGridState();
}

class _SfDataGridState extends State<SfDataGrid> {
  Map<String, GridCellRendererBase> _cellRenderers = {};
  _RowGenerator _rowGenerator;
  _VisualContainerHelper _container;
  _DataGridStateDetails _dataGridStateDetails;
  _DataGridSettings _dataGridSettings;
  ColumnSizer _columnSizer;
  TextDirection _textDirection = TextDirection.ltr;
  SfDataGridThemeData _dataGridThemeData;
  static const double _minWidth = 300.0;
  static const double _minHeight = 300.0;
  DataGridSource _source;
  List<GridColumn> _columns = [];
  SelectionManagerBase _rowSelectionManager;
  DataGridController _controller;
  _CurrentCellManager _currentCell;

  set textDirection(TextDirection newTextDirection) {
    if (this._textDirection == newTextDirection) return;
    this._textDirection = newTextDirection;
    this._dataGridSettings.textDirection = newTextDirection;
    this._container._needToSetHorizontalOffset = true;
  }

  set dataGridThemeData(SfDataGridThemeData newThemeData) {
    if (this._dataGridThemeData == newThemeData) return;

    var canUpdate = this._dataGridThemeData != null &&
        this._dataGridThemeData != newThemeData;
    this._dataGridThemeData = newThemeData;
    this._dataGridSettings.dataGridThemeData = newThemeData;
    this._updateDecoration();
    if (canUpdate) this._container._refreshViewStyle();
  }

  @override
  void initState() {
    this._dataGridSettings = _DataGridSettings();
    this._dataGridStateDetails = this._getPropertyValues;
    this._dataGridSettings.gridPaint = Paint();
    this._rowGenerator =
        _RowGenerator(dataGridStateDetails: this._dataGridStateDetails);
    this._container = _VisualContainerHelper(rowGenerator: this._rowGenerator);
    this._setUp();
    this._updateDataGridStateDetails();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    SyncfusionLicense.validateLicense(context);
    if (kIsWeb) this._updateBoxPainter();
    return LayoutBuilder(builder: (_context, _constraints) {
      final double _measuredHeight = this._dataGridSettings.viewHeight =
          _constraints.maxHeight.isInfinite
              ? _minHeight
              : _constraints.maxHeight;
      final double _measuredWidth = this._dataGridSettings.viewWidth =
          _constraints.minWidth.isInfinite ? _minWidth : _constraints.maxWidth;

      if (!this._container._isGridLoaded) {
        this._gridLoaded();
        if (this._textDirection == TextDirection.rtl)
          this._container._needToSetHorizontalOffset = true;
        this._container._isDirty = true;
        this._columnSizer._isColumnSizerLoadedInitially = true;
      }

      return _ScrollViewWidget(
        dataGridStateDetails: _dataGridStateDetails,
        width: _measuredWidth,
        height: _measuredHeight,
      );
    });
  }

  void _setUp() {
    this._initializeDataGridDataSource();
    this._initializeCellRendererCollection();

    //DataGrid Controller
    this._controller = this._dataGridSettings.controller =
        widget.controller ?? DataGridController();
    this._controller._dataGridStateDetails = this._dataGridStateDetails;

    //AutoFit controller initializing
    this._columnSizer = widget.columnSizer ?? ColumnSizer();
    this._columnSizer._dataGridStateDetails = this._dataGridStateDetails;

    //CurrentCell Manager initializing
    this._currentCell = this._dataGridSettings.currentCell =
        _CurrentCellManager(this._dataGridStateDetails);

    //Selection Manager initializing
    this._rowSelectionManager = this._dataGridSettings.rowSelectionManager =
        widget.selectionManager ??
            RowSelectionManager(
                dataGridStateDetails: this._dataGridStateDetails);
    this._rowSelectionManager?._dataGridStateDetails =
        this._dataGridStateDetails;
    this
        ._controller
        ?.addListener(this._rowSelectionManager.handleDataGridSourceChanges);

    this._initializeProperties();
  }

  @protected
  void _gridLoaded() {
    this._container._refreshDefaultLineSize();
    this._refreshContainerAndView();
  }

  @protected
  void _refreshContainerAndView({bool isDataSourceChanged = false}) {
    if (this._rowGenerator == null) {
      return;
    }

    if (isDataSourceChanged)
      this
          ._rowSelectionManager
          ._updateSelectionController(isDataSourceChanged: isDataSourceChanged);

    this._ensureSelectionProperties();
    this._container._refreshHeaderLineCount();
    this._container._updateRowAndColumnCount();
    this._container._isGridLoaded = true;
  }

  void _initializeDataGridDataSource() {
    if (this._source != widget.source) {
      this._source?.removeListener(_handleDataGridSourceChanged);
      this._source = widget.source;
      this._source?.addListener(_handleDataGridSourceChanged);
    }
  }

  void _initializeProperties() {
    if (!listEquals<GridColumn>(this._columns, widget.columns)) {
      this._columns.clear();
      this._columns.addAll(widget.columns ?? this._columns);
    }

    this._rowSelectionManager?._dataGridStateDetails =
        this._dataGridStateDetails;
    this._currentCell?._dataGridStateDetails = this._dataGridStateDetails;
    this._columnSizer?._dataGridStateDetails = this._dataGridStateDetails;
    this._updateDataGridStateDetails();
  }

  void _initializeCellRendererCollection() {
    this._cellRenderers = {};
    this._cellRenderers['TextField'] =
        GridCellTextFieldRenderer(this._dataGridStateDetails);
    this._cellRenderers['ColumnHeader'] =
        GridHeaderCellRenderer(this._dataGridStateDetails);
    this._cellRenderers['Numeric'] =
        GridCellNumericTextFieldRenderer(this._dataGridStateDetails);
    this._cellRenderers['Widget'] =
        GridCellWidgetRenderer(this._dataGridStateDetails);
    this._cellRenderers['DateTime'] =
        GridCellDateTimeRenderer(this._dataGridStateDetails);

    if (widget.onCellRenderersCreated != null) {
      widget.onCellRenderersCreated(this._cellRenderers);
      for (var renderer in this._cellRenderers.entries) {
        renderer.value._dataGridStateDetails = this._dataGridStateDetails;
      }
    }
  }

  void _processCellUpdate(RowColumnIndex rowColumnIndex) {
    if (rowColumnIndex != RowColumnIndex(-1, -1)) {
      var rowIndex = _GridIndexResolver.resolveToRowIndex(
          this._dataGridSettings, rowColumnIndex.rowIndex);
      var columnIndex = _GridIndexResolver.resolveToScrollColumnIndex(
          this._dataGridSettings, rowColumnIndex.columnIndex);

      var dataRow = this._rowGenerator.items.firstWhere(
          (dataRow) => dataRow.rowIndex == rowIndex,
          orElse: () => null);

      if (dataRow == null) return;

      var dataCell = dataRow._visibleColumns.firstWhere(
          (dataCell) => dataCell.columnIndex == columnIndex,
          orElse: () => null);

      if (dataCell == null) {
        dataRow = null;
        return;
      }

      this.setState(() {
        dataCell._isDirty = true;
        dataCell._updateColumn();
        dataCell = null;
      });

      dataCell = null;
      dataRow = null;
    }
  }

  void _processUpdateDataSource() {
    this.setState(() {
      this._initializeDataGridDataSource();
      this._dataGridSettings.source = this._source;
      if (!listEquals<GridColumn>(this._columns, widget.columns)) {
        if (widget.selectionMode != SelectionMode.none &&
            widget.navigationMode == GridNavigationMode.cell)
          this
              ._rowSelectionManager
              ._onRowColumnChanged(-1, widget.columns.length);
        this._resetColumn();
      }

      if (widget.selectionMode != SelectionMode.none &&
          widget.navigationMode == GridNavigationMode.cell)
        this
            ._rowSelectionManager
            ._onRowColumnChanged(widget.source.dataSource.length, -1);
      this._container._updateRowAndColumnCount();
      this._container._refreshView();
      this._container._isDirty = true;
    });
  }

  void _resetColumn() {
    this._columns.clear();
    this._columns.addAll(widget.columns);
    this._dataGridSettings.columns = this._columns;
    this._rowGenerator.items.forEach((dataRow) =>
        dataRow._visibleColumns.forEach((col) => col.columnIndex = -1));
    if (_textDirection == TextDirection.rtl)
      this._container._needToSetHorizontalOffset = true;
    this._container._needToRefreshColumn = true;
  }

  void _handleDataGridSourceChanged(
      {RowColumnIndex rowColumnIndex, String propertyName}) {
    if (rowColumnIndex != null && propertyName == null) {
      this._processCellUpdate(rowColumnIndex);
    }

    if (rowColumnIndex == null && propertyName == null) {
      this._processUpdateDataSource();
    }
  }

  void _updateDataGridStateDetails() {
    this._dataGridSettings
      ..textDirection = this._textDirection
      ..cellRenderers = this._cellRenderers
      ..container = this._container
      ..rowGenerator = this._rowGenerator
      ..rowHeight = widget.rowHeight
      ..headerRowHeight = widget.headerRowHeight
      ..source = this._source
      ..columns = this._columns
      ..defaultColumnWidth = widget.defaultColumnWidth
      ..headerLineCount = this._container._headerLineCount
      ..onQueryCellStyle = widget.onQueryCellStyle
      ..onQueryRowStyle = widget.onQueryRowStyle
      ..cellBuilder = widget.cellBuilder
      ..onQueryRowHeight = widget.onQueryRowHeight
      ..dataGridThemeData = this._dataGridThemeData
      ..gridLinesVisibility = widget.gridLinesVisibility
      ..columnWidthMode = widget.columnWidthMode
      ..columnSizer = this._columnSizer
      ..columnWidthCalculationMode = widget.columnWidthCalculationMode
      ..columnWidthCalculationRange = widget.columnWidthCalculationRange
      ..selectionMode = widget.selectionMode
      ..onSelectionChanged = widget.onSelectionChanged
      ..onSelectionChanging = widget.onSelectionChanging
      ..navigationMode = widget.navigationMode
      ..onCurrentCellActivated = widget.onCurrentCellActivated
      ..onCurrentCellActivating = widget.onCurrentCellActivating;
  }

  _DataGridSettings _getPropertyValues() {
    return this._dataGridSettings;
  }

  void _updateProperties(SfDataGrid oldWidget) {
    var isSourceChanged = widget.source != oldWidget.source;
    var isDataSourceChanged = !listEquals<Object>(
        oldWidget.source.dataSource, widget.source.dataSource);
    var isColumnsChanged =
        !listEquals<GridColumn>(this._columns, widget.columns);
    var isSelectionManagerChanged =
        oldWidget.selectionManager != widget.selectionManager ||
            oldWidget.selectionMode != widget.selectionMode;
    var isColumnSizerChanged = oldWidget.columnSizer != widget.columnSizer ||
        oldWidget.columnWidthCalculationMode !=
            widget.columnWidthCalculationMode ||
        oldWidget.columnWidthMode != widget.columnWidthMode ||
        oldWidget.columnWidthCalculationRange !=
            widget.columnWidthCalculationRange;
    var isDataGridControllerChanged = oldWidget.controller != widget.controller;

    if (isSourceChanged ||
        isColumnsChanged ||
        isDataSourceChanged ||
        isSelectionManagerChanged ||
        isColumnSizerChanged ||
        isDataGridControllerChanged ||
        oldWidget.rowHeight != widget.rowHeight ||
        oldWidget.headerRowHeight != widget.headerRowHeight ||
        oldWidget.defaultColumnWidth != widget.defaultColumnWidth ||
        oldWidget.navigationMode != widget.navigationMode) {
      if (isSourceChanged) this._initializeDataGridDataSource();

      if (isDataGridControllerChanged) {
        this._controller = this._dataGridSettings.controller =
            widget.controller ?? this._controller;
        this._controller._dataGridStateDetails = this._dataGridStateDetails;
      }

      this._initializeProperties();
      this._container._refreshDefaultLineSize();

      this._updateSelectionController(
          oldWidget: oldWidget,
          isDataGridControlChanged: isDataGridControllerChanged,
          isSelectionManagerChanged: isSelectionManagerChanged,
          isSourceChanged: isSourceChanged,
          isDataSourceChanged: isDataSourceChanged);

      this._container._updateRowAndColumnCount();

      if (isSourceChanged || isColumnsChanged || isColumnSizerChanged) {
        this._rowGenerator.items.forEach((dataRow) =>
            dataRow._visibleColumns.forEach((col) => col.columnIndex = -1));
        this._container._needToRefreshColumn = true;
      }

      if (isSourceChanged || isDataSourceChanged) {
        this._container._refreshView();
      }

      this._container._isDirty = true;
    } else {
      if (oldWidget.gridLinesVisibility != widget.gridLinesVisibility) {
        this._initializeProperties();
        this._container._isDirty = true;
      }
    }
  }

  void _updateSelectionController(
      {SfDataGrid oldWidget,
      bool isSelectionManagerChanged,
      bool isDataGridControlChanged,
      bool isSourceChanged,
      bool isDataSourceChanged}) {
    if (isSourceChanged) {
      oldWidget.controller?.removeListener(
          this._rowSelectionManager.handleDataGridSourceChanges);
      widget.controller
          ?.addListener(this._rowSelectionManager.handleDataGridSourceChanges);
    }

    if (isSelectionManagerChanged) {
      this._rowSelectionManager = this._dataGridSettings.rowSelectionManager =
          widget.selectionManager ?? this._rowSelectionManager;

      this._rowSelectionManager._dataGridStateDetails =
          this._dataGridStateDetails;
    }

    this._rowSelectionManager?._updateSelectionController(
        isSelectionModeChanged: oldWidget.selectionMode != widget.selectionMode,
        isNavigationModeChanged:
            oldWidget.navigationMode != widget.navigationMode,
        isDataSourceChanged: isDataSourceChanged);

    if (isDataGridControlChanged) this._ensureSelectionProperties();
  }

  void _ensureSelectionProperties() {
    if (this._controller.selectedRows.isNotEmpty) {
      this._rowSelectionManager?.onSelectedRowsChanged();
    }

    if (this._controller.selectedRow != null) {
      this._rowSelectionManager?.onSelectedRowChanged();
    }

    if (this._controller.selectedIndex != null &&
        this._controller.selectedIndex != -1) {
      this._rowSelectionManager?.onSelectedIndexChanged();
    }
  }

  void _updateBoxPainter() {
    if (widget.selectionMode == SelectionMode.multiple &&
        widget.navigationMode == GridNavigationMode.row) {
      if (this._dataGridSettings.configuration == null)
        this._dataGridSettings.configuration =
            createLocalImageConfiguration(context);
      if (this._dataGridSettings.boxPainter == null) this._updateDecoration();
    }
  }

  void _updateDecoration() {
    final borderSide = BorderSide(
        color: this._dataGridThemeData.currentCellStyle.borderColor,
        width: 1.0);
    final decoration = BoxDecoration(
        border: Border(
            bottom: borderSide,
            top: borderSide,
            left: borderSide,
            right: borderSide));
    this._dataGridSettings.boxPainter = decoration.createBoxPainter();
  }

  @override
  void didChangeDependencies() {
    this.textDirection = Directionality.of(context);
    this.dataGridThemeData = SfDataGridTheme.of(context);
    super.didChangeDependencies();
  }

  @override
  void didUpdateWidget(SfDataGrid oldWidget) {
    super.didUpdateWidget(oldWidget);
    this._updateProperties(oldWidget);
  }

  @override
  void dispose() {
    this._source?.removeListener(_handleDataGridSourceChanged);
    this._dataGridSettings?.gridPaint ??= null;
    this._dataGridSettings?.boxPainter ??= null;
    this._dataGridSettings?.configuration ??= null;
    this._dataGridThemeData = null;
    super.dispose();
  }
}

class _DataGridSettings {
  Map<String, GridCellRendererBase> cellRenderers;
  DataGridSource source;
  List<GridColumn> columns;
  double rowHeight;
  double headerRowHeight;
  double defaultColumnWidth;
  _VisualContainerHelper container;
  _RowGenerator rowGenerator;
  int headerLineCount;
  SfDataGridThemeData dataGridThemeData;
  CellBuilderCallback cellBuilder;
  QueryCellStyleCallback onQueryCellStyle;
  QueryRowStyleCallback onQueryRowStyle;
  QueryRowHeightCallback onQueryRowHeight;
  TextDirection textDirection;
  GridLinesVisibility gridLinesVisibility;
  ColumnWidthMode columnWidthMode;
  ColumnSizer columnSizer;
  ColumnWidthCalculationMode columnWidthCalculationMode;
  ColumnWidthCalculationRange columnWidthCalculationRange;
  SelectionManagerBase rowSelectionManager;
  SelectionMode selectionMode;
  SelectionChangingCallback onSelectionChanging;
  SelectionChangedCallback onSelectionChanged;
  DataGridController controller;
  _CurrentCellManager currentCell;
  GridNavigationMode navigationMode;
  CurrentCellActivatedCallback onCurrentCellActivated;
  CurrentCellActivatingCallback onCurrentCellActivating;
  ScrollController verticalController;
  ScrollController horizontalController;
  FocusNode dataGridFocusNode;
  double viewWidth;
  double viewHeight;
  Paint gridPaint;
  ImageConfiguration configuration;
  BoxPainter boxPainter;
}
